# (C) Copyright 2024-2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from functools import partial
from pathlib import Path
from typing import Annotated

from pydantic import AfterValidator
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field
from pydantic import NonNegativeInt

from anemoi.utils.schemas import BaseModel
from anemoi.utils.schemas.errors import allowed_values


class HardwareSchema(BaseModel):
    accelerator: Annotated[
        str,
        AfterValidator(partial(allowed_values, values=["cpu", "gpu", "auto", "cuda", "tpu"])),
    ] = "auto"
    "Accelerator to use for training."
    num_gpus_per_node: NonNegativeInt = 1
    "Number of GPUs per node."
    num_nodes: NonNegativeInt = 1
    "Number of nodes."
    num_gpus_per_model: NonNegativeInt = 1
    "Number of GPUs per model."
    num_gpus_per_ensemble: NonNegativeInt = 1
    "Number of GPUs per ensemble."


class InputSchema(PydanticBaseModel):
    dataset: Path | dict[str, Path] | None = Field(default=None)  # dict option for multiple datasets
    "Path to the dataset file."
    graph: Path | None = None
    "Path to the graph file."
    truncation: Path | None = None
    "Path to the truncation matrix file."
    truncation_inv: Path | None = None
    "Path to the inverse truncation matrix file."
    loss_matrices_path: Path | None = None
    "Path to the directory containing smoothing matrices for the multiscale loss"
    warm_start: Path | None = None
    "Path of the checkpoint file to use for warm starting the training"


class CheckpointsSchema(BaseModel):
    root: Path
    "Root directory for saving checkpoint files."
    every_n_epochs: str = "anemoi-by_epoch-epoch_{epoch:03d}-step_{step:06d}"
    "File name pattern for checkpoint files saved by epoch frequency."
    every_n_train_steps: str = "anemoi-by_step-epoch_{epoch:03d}-step_{step:06d}"
    "File name pattern for checkpoint files saved by step frequency."
    every_n_minutes: str = "anemoi-by_time-epoch_{epoch:03d}-step_{step:06d}"
    "File name pattern for checkpoint files saved by time frequency (minutes)."


class Logs(PydanticBaseModel):
    root: Path
    wandb: Path | None = None
    "Path to output wandb logs."
    mlflow: Path | None = None
    "Path to output mlflow logs."
    tensorboard: Path | None = None
    "Path to output tensorboard logs."


class OutputSchema(BaseModel):
    root: Path | None = None
    "Path to the output directory."
    logs: Logs | None = None
    "Logging directories."
    checkpoints: CheckpointsSchema = Field(default_factory=CheckpointsSchema)
    "Paths to the checkpoints."
    plots: Path | None = None
    "Path to the plots directory."
    profiler: Path | None
    "Path to the profiler directory."


class SystemSchema(BaseModel):
    hardware: HardwareSchema
    "Specification of hardware and compute resources available including the number of nodes, GPUs, and accelerator."
    input: InputSchema
    "Definitions of specific input and output artifacts used relative to the directories defined in `output`."
    output: OutputSchema
    "High-level directory structure describing where data is read from."



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import pytest
import torch
from hydra.errors import InstantiationException
from omegaconf import DictConfig

from anemoi.training.losses import CombinedLoss
from anemoi.training.losses import MAELoss
from anemoi.training.losses import MSELoss
from anemoi.training.losses import get_loss_function


def test_combined_loss() -> None:
    """Test the combined loss function."""
    loss = get_loss_function(
        DictConfig(
            {
                "_target_": "anemoi.training.losses.CombinedLoss",
                "losses": [
                    {"_target_": "anemoi.training.losses.MSELoss"},
                    {"_target_": "anemoi.training.losses.MAELoss"},
                ],
                "scalers": ["test"],
                "loss_weights": [1.0, 0.5],
            },
        ),
        scalers={"test": (-1, torch.ones(2))},
    )
    assert isinstance(loss.losses[0], MSELoss)
    assert "test" in loss.losses[0].scaler

    assert isinstance(loss.losses[1], MAELoss)
    assert "test" in loss.losses[1].scaler


def test_combined_loss_invalid_loss_weights() -> None:
    """Test the combined loss function with invalid loss weights."""
    with pytest.raises(InstantiationException):
        get_loss_function(
            DictConfig(
                {
                    "_target_": "anemoi.training.losses.combined.CombinedLoss",
                    "losses": [
                        {"_target_": "anemoi.training.losses.MSELoss"},
                        {"_target_": "anemoi.training.losses.MAELoss"},
                    ],
                    "scalers": ["test"],
                    "loss_weights": [1.0, 0.5, 1],
                },
            ),
            scalers={"test": (-1, torch.ones(2))},
        )


def test_combined_loss_equal_weighting() -> None:
    """Test equal weighting when not given."""
    loss = get_loss_function(
        DictConfig(
            {
                "_target_": "anemoi.training.losses.CombinedLoss",
                "losses": [
                    {"_target_": "anemoi.training.losses.MSELoss"},
                    {"_target_": "anemoi.training.losses.MAELoss"},
                ],
            },
        ),
        scalers={},
    )
    assert all(weight == 1.0 for weight in loss.loss_weights)


def test_combined_loss_seperate_scalers() -> None:
    """Test that scalers are passed to the correct loss function."""
    loss = get_loss_function(
        DictConfig(
            {
                "_target_": "anemoi.training.losses.CombinedLoss",
                "losses": [
                    {"_target_": "anemoi.training.losses.MSELoss", "scalers": ["test"]},
                    {"_target_": "anemoi.training.losses.MAELoss", "scalers": ["test2"]},
                ],
                "scalers": ["test", "test2"],
                "loss_weights": [1.0, 0.5],
            },
        ),
        scalers={"test": (-1, torch.ones(2)), "test2": (-1, torch.ones(2))},
    )
    assert isinstance(loss, CombinedLoss)

    assert isinstance(loss.losses[0], MSELoss)
    assert "test" in loss.losses[0].scaler
    assert "test2" not in loss.losses[0].scaler

    assert isinstance(loss.losses[1], MAELoss)
    assert "test" not in loss.losses[1].scaler
    assert "test2" in loss.losses[1].scaler



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from collections import defaultdict

from hydra.utils import instantiate
from omegaconf import DictConfig
from omegaconf import OmegaConf

from anemoi.models.data_indices.collection import IndexCollection
from anemoi.models.data_indices.tensor import OutputTensorIndex
from anemoi.training.losses.base import BaseLoss
from anemoi.training.losses.scaler_tensor import TENSOR_SPEC
from anemoi.training.utils.variables_metadata import ExtractVariableGroupAndLevel

METRIC_RANGE_DTYPE = dict[str, list[int]]

NESTED_LOSSES = ["anemoi.training.losses.MultiscaleLossWrapper"]
LOGGER = logging.getLogger(__name__)


# Future import breaks other type hints TODO Harrison Cook
def get_loss_function(
    config: DictConfig,
    scalers: dict[str, TENSOR_SPEC] | None = None,
    data_indices: dict | None = None,
    **kwargs,
) -> BaseLoss:
    """Get loss functions from config.

    Can be ModuleList if multiple losses are specified.

    Parameters
    ----------
    config : DictConfig
        Loss function configuration, should include `scalers` if scalers are to be added to the loss function.
    scalers : TENSOR_SPEC, optional,
        Scalers which can be added to the loss function. Defaults to None., by default None
        If a scaler is to be added to the loss, ensure it is in `scalers` in the loss config.
        For instance, if `scalers: ['variable']` is set in the config, and `variable` in `scalers`
        `variable` will be added to the scaler of the loss function.
    data_indices : dict, optional
        Indices of the training data
    kwargs : Any
        Additional arguments to pass to the loss function

    Returns
    -------
    BaseLoss | torch.nn.ModuleDict
        The loss function, or dict of metrics, to use for training/validation.

    Raises
    ------
    TypeError
        If not a subclass of `BaseLoss`.
    ValueError
        If scaler is not found in valid scalers
    """
    loss_config = OmegaConf.to_container(config, resolve=True)
    scalers_to_include = loss_config.pop("scalers", [])

    if "_target_" in loss_config and loss_config["_target_"] in NESTED_LOSSES:
        per_scale_loss_config = loss_config.pop("per_scale_loss")
        per_scale_loss = get_loss_function(OmegaConf.create(per_scale_loss_config), scalers, data_indices)
        return instantiate(loss_config, per_scale_loss=per_scale_loss, **kwargs)

    if scalers is None:
        scalers = {}

    if "*" in scalers_to_include:
        scalers_to_include = [s for s in list(scalers.keys()) if f"!{s}" not in scalers_to_include]

    loss_function = instantiate(loss_config, **kwargs, _recursive_=False)

    if not isinstance(loss_function, BaseLoss):
        error_msg = f"Loss must be a subclass of 'BaseLoss', not {type(loss_function)}"
        raise TypeError(error_msg)
    _apply_scalers(loss_function, scalers_to_include, scalers, data_indices)

    return loss_function


def _apply_scalers(
    loss_function: BaseLoss,
    scalers_to_include: list,
    scalers: dict[str, TENSOR_SPEC] | None,
    data_indices: dict | None,
) -> None:
    """Attach scalers to a loss function and set data indices if needed."""
    for key in scalers_to_include:
        if key not in scalers or []:
            error_msg = f"Scaler {key!r} not found in valid scalers: {list(scalers.keys())}"
            raise ValueError(error_msg)
        if key in ["stdev_tendency", "var_tendency"]:
            for var_key, idx in data_indices.model.output.name_to_index.items():
                if idx in data_indices.model.output.prognostic and data_indices.data.output.name_to_index.get(
                    var_key,
                ):
                    scaling = scalers[key][1][idx]
                    LOGGER.info("Parameter %s is being scaled by statistic_tendencies by %.2f", var_key, scaling)
        loss_function.add_scaler(*scalers[key], name=key)

        if hasattr(loss_function, "set_data_indices"):
            loss_function.set_data_indices(data_indices)


def _get_metric_ranges(
    extract_variable_group_and_level: ExtractVariableGroupAndLevel,
    output_data_indices: OutputTensorIndex,
    metrics_to_log: list,
) -> METRIC_RANGE_DTYPE:
    metric_ranges = defaultdict(list)

    for key, idx in output_data_indices.name_to_index.items():
        variable_group, variable_ref, _ = extract_variable_group_and_level.get_group_and_level(key)

        # Add metrics for grouped variables and variables in default group
        metric_ranges[f"{variable_group}_{variable_ref}"].append(idx)

        # Specific metrics from hydra to log in logger
        if key in metrics_to_log:
            metric_ranges[key] = [idx]

    # Add the full list of output indices
    metric_ranges["all"] = output_data_indices.full.tolist()
    return metric_ranges


def get_metric_ranges(
    config: DictConfig,
    data_indices: IndexCollection,
    metadata_extractor: ExtractVariableGroupAndLevel,
) -> tuple[METRIC_RANGE_DTYPE, METRIC_RANGE_DTYPE]:

    metrics_to_log = config.training.metrics or []

    return _get_metric_ranges(
        metadata_extractor,
        data_indices.model.output,
        metrics_to_log,
    )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import numpy as np
import pandas as pd
import torch
import torch.distributed as dist
from pytorch_lightning.callbacks import TQDMProgressBar
from pytorch_lightning.profilers import Profiler
from pytorch_lightning.profilers import PyTorchProfiler
from pytorch_lightning.profilers import SimpleProfiler
from pytorch_lightning.utilities import rank_zero_only

if TYPE_CHECKING:
    import importlib

    import pytorch_lightning as pl
    from omegaconf import DictConfig
    from pytorch_lightning.utilities.types import STEP_OUTPUT

    from anemoi.training.train.tasks.base import BaseGraphModule

    if importlib.util.find_spec("ipywidgets") is not None:
        from tqdm.auto import tqdm as _tqdm
    else:
        from tqdm import tqdm as _tqdm

from torch.profiler import profile

from anemoi.training.diagnostics.mlflow.logger import AnemoiMLflowLogger

LOGGER = logging.getLogger(__name__)


def check_torch_version() -> bool:
    torch_version = torch.__version__
    version_nums = torch_version.split(".")
    major_version = int(version_nums[0])
    minor_version = int(version_nums[1])
    if major_version == 2 and minor_version >= 1:
        return True
    LOGGER.error("Memory snapshot is only supported for torch >= 2.1")
    return False


def convert_to_seconds(time_str: str) -> float:
    import re

    pattern = r"(\d+(\.\d+)?)\s*([a-zA-Z]+)"
    # Use regex to find matches
    match = re.match(pattern, time_str)

    # Check if match is found
    if match:
        # Extract numeric part and unit part
        numeric_part = float(match.group(1))
        unit = match.group(3)

        # Convert the unit to seconds
        if unit == "s":
            return numeric_part
        if unit == "ds":
            return numeric_part / 10  # Convert decaseconds to seconds
        if unit == "cs":
            return numeric_part / 100  # Convert centiseconds to seconds
        if unit == "ms":
            return numeric_part / 1000  # Convert milliseconds to seconds
        error_msg = (
            "Invalid unit. Supported units are: 's' (seconds)'"
            "'ds' (decaseconds), 'cs' (centiseconds) and 'ms' (miliseconds) .",
        )
        raise ValueError(error_msg)
    error_msg = "Invalid time format. The time should be in the format: 'numeric_part unit'. For example: '10 ms'"
    raise ValueError(error_msg)


PROFILER_ACTIONS = [
    r"\[Strategy]\w+\.batch_to_device",
    r"\[Strategy]\w+\.backward",
    r"\[Strategy]\w+\.training_step",
    r"\[Strategy]\w+\.validation_step",
    r"\[Strategy]\w+\.batch_to_device",
    "run_training_epoch",
    "run_training_batch",
    r"\[_EvaluationLoop\]\.\w+",
    r"\[_TrainingEpochLoop\]\.\w+",
    r"\[LightningDataModule]\w+\.train_dataloader",
    r"\[LightningDataModule]\w+\.val_dataloader",
    r"\[LightningDataModule]\w+\.state_dict",
    r"\[LightningDataModule]\w+\.setup",
    r"\[LightningDataModule]\w+\.prepare_data",
    r"\[LightningDataModule]\w+\.teardown",
    r"\[LightningModule]\w+\.optimizer_step",
    r"\[LightningModule]\w+\.configure_gradient_clipping",
    r"\[LightningModule]\w+\.on_validation_model_eval",
    r"\[LightningModule]\w+\.optimizer_zero_grad",
    r"\[LightningModule]\w+\.transfer_batch_to_device",
    r"\[LightningModule]\w+\.on_validation_model_train",
    r"\[LightningModule]\w+\.configure_optimizers",
    r"\[LightningModule]\w+\.lr_scheduler_step",
    r"\[LightningModule]\w+\.configure_sharded_model",
    r"\[LightningModule]\w+\.setup",
    r"\[LightningModule]\w+\.prepare_data",
    r"\[Callback\](.*Plot*)",
    r"\[Callback\](.*Checkpoint*)",
]

GPU_METRICS_DICT = {
    "GPU device utilization (%)": "gpu",
    "GPU memory use (%)": "memory",
    "GPU memory allocated (%)": "memoryAllocated",
    "GPU memory allocated (GB)": "memoryAllocatedBytes",
}


class WandBSystemSummarizer:
    """Summarize System Metrics provided by W&B logger."""

    def __init__(self, wandb_logger: pl.loggers.WandbLogger):

        run_dict = wandb_logger._wandb_init
        self.run_id_path = f"{run_dict['entity']}/{run_dict['project']}/{run_dict['id']}"

    def get_wandb_metrics(self) -> (pd.DataFrame, dict):
        """Fetches system metrics and metadata from a W&B run."""
        import wandb

        run = wandb.Api().run(self.run_id_path)
        system_metrics = run.history(stream="events")
        metadata_dict = run.metadata
        system_metrics = system_metrics.dropna()
        return system_metrics, metadata_dict

    def summarize_gpu_metrics(self, df: pd.DataFrame) -> dict[str, float]:
        """Given the System Metrics DataFrame, summarized the GPU metrics.

        - gpu.{gpu_index}.memory - GPU memory utilization in percent for each GPU
        - gpu.{gpu_index}.memoryAllocated - GPU memory allocated as % of the total available memory for each GPU
        - gpu.{gpu_index}.memoryAllocatedBytes - GPU memory allocated in bytes for each GPU
        - gpu.{gpu_index}.gpu - GPU utilization in percent for each GPU
        """
        average_metric = {}
        col_names = df.columns
        for gpu_metric_name, gpu_metric in GPU_METRICS_DICT.items():
            pattern = rf"system.gpu.\d.{gpu_metric}$"
            sub_gpu_cols = [string for string in col_names if re.match(pattern, string)]
            metrics_per_gpu = df[sub_gpu_cols].mean(axis=0)
            if gpu_metric == "memoryAllocatedBytes":
                metrics_per_gpu = metrics_per_gpu * 1e-9
            average_metric[gpu_metric_name] = metrics_per_gpu.mean()
            # Just add metrics per gpu to the report if we have more than 1 GPU
            if metrics_per_gpu.shape[0] > 1:
                metrics_per_gpu.index = ["   " + index for index in metrics_per_gpu.index]
                average_metric.update(dict(metrics_per_gpu))
        return average_metric

    def summarize_system_metrics(self) -> dict[str, float]:
        r"""Summarizes the System metrics from a W&B run.

        Some of the metrics included are:
        - cpu.{}.cpu_percent - CPU usage of the system on a per-core basis.
        - system.memory - Represents the total system memory usage as a percentage of the total available memory.
        - system.cpu - Percentage of CPU usage by the process, normalized by the number of available CPUs
        - system.disk.\\.usageGB - (Represents the total system disk usage in gigabytes (GB))
        - system.proc.memory.percent - Indicates the memory usage of the process as a % of the total available memory

        More information about W&B system metrics can be found here:
        https://docs.wandb.ai/guides/app/features/system-metrics
        """
        system_metrics_df, metadata_dict = self.get_wandb_metrics(self.run_id_path)

        col_names = system_metrics_df.columns
        system_metrics = {}

        n_cpus = metadata_dict["cpu_count"]
        cpu_cols = list(filter(lambda k: "cpu." in k, col_names))
        system_metrics["avg CPU usage (%)"] = (system_metrics_df[cpu_cols].sum(axis=1) / n_cpus).mean()

        system_metrics_gpu = self.summarize_gpu_metrics(system_metrics_df)
        system_metrics.update(system_metrics_gpu)

        system_metrics["avg Memory usage (%)"] = system_metrics_df["system.memory"].mean()
        system_metrics["avg Disk usage (GB)"] = system_metrics_df["system.disk.\\.usageGB"].mean()
        system_metrics["avg Disk usage  (%)"] = system_metrics_df["system.disk.\\.usagePercent"].mean()

        system_metrics["execution time (sec)"] = system_metrics_df["_runtime"].iloc[-1]  # in seconds
        return system_metrics


class MLFlowSystemSummarizer:
    """Summarize System Metrics provided by MlFlow logger."""

    def __init__(self, mlflow_logger: pl.loggers.MLFlowLogger):
        self.run_id = mlflow_logger.run_id
        self.mlflow_client = mlflow_logger._mlflow_client

    @property
    def system_metrics(self) -> list[str]:
        run = self.mlflow_client.get_run(self.run_id)
        return [metric for metric in run.data.metrics if "system" in metric]

    def _clean_metric_name(self, metric_name: str) -> str:
        return (
            metric_name.replace("system.", "avg ")
            .replace("_", " ")
            .replace("megabytes", "MB")
            .replace("percentage", "%")
        )

    def _get_mean(self, pattern: str, df: pd.DataFrame) -> float:
        # Filter rows containing the pattern in the 'metric' column
        filtered_rows = df[df["metric"].str.contains(pattern)]
        return filtered_rows.loc[:, "value"].astype(np.float32).mean()

    def _extract_gpu_metrics(self, df: pd.DataFrame) -> pd.DataFrame:
        # Define the pattern you want to search for
        pattern = r"gpu\s\d+\s+utilization"
        df.loc[len(df.index)] = ["avg GPU utilization (%)", self._get_mean(pattern, df)]

        pattern = r"gpu\s\d+\s+memory\s+usage\s+%"
        df.loc[len(df.index)] = ["avg GPU memory usage %", self._get_mean(pattern, df)]

        pattern = r"gpu\s\d+\s+memory\s+usage\s+MB"
        df.loc[len(df.index)] = ["avg GPU memory usage MB", self._get_mean(pattern, df)]

        return df

    def summarize_mlflow_system_metrics(self) -> pd.DataFrame:
        rows = []
        for metric in self.system_metrics:
            metric = self.mlflow_client.get_metric_history(self.run_id, metric)
            avg_value = sum(m.value for m in metric) / len(metric)
            metric_name = self._clean_metric_name(metric[0].key)
            rows.append({"metric": metric_name, "value": f"{avg_value:.2f}"})
        return self._extract_gpu_metrics(pd.DataFrame(rows))


class DummyProfiler(Profiler):
    """Placeholder profiler."""

    def __init__(self):
        super().__init__()

    def start(self, *args, **kwargs) -> None:
        pass

    def stop(self, *args, **kwargs) -> None:
        pass


def _convert_npint_to_int(obj: Any) -> dict | list | int | str | float:
    """Recursively converts all np.int64 values in the input to Python int."""
    # Recursively converts all np.int64 to int
    if isinstance(obj, dict):
        return {k: _convert_npint_to_int(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [_convert_npint_to_int(item) for item in obj]
    if isinstance(obj, np.integer):
        return int(obj)  # Convert np.int64 to int
    return obj


class PatchedProfile(profile):

    def _get_distributed_info(self) -> dict[str, str]:
        dist_info = super()._get_distributed_info()
        return _convert_npint_to_int(dist_info)


class BenchmarkProfiler(Profiler):
    """Custom PyTorch Lightning profiler for benchmarking."""

    def __init__(self, config: DictConfig) -> None:
        super().__init__(config)

        self.config = config
        self.warmup = self.config.diagnostics.benchmark_profiler.memory.warmup
        if not self.warmup:
            self.warmup = 0
        self.num_steps = self.config.diagnostics.benchmark_profiler.memory.steps

        if self.config.diagnostics.benchmark_profiler.memory.extra_plots:
            assert (
                self.num_steps <= self.config.training.num_sanity_val_steps
            ), "Sanity steps should be less than snapshot steps, to avoid memory issues"

        self.dirpath = None
        self.create_output_path()
        # the profilers need to be initialised before the setup method because
        # actions like configuring callbacks would trigger the profiler
        self.memory_profiler = DummyProfiler  # dummy profiler to be used as placeholder
        self.time_profiler = DummyProfiler  # dummy profiler to be used as placeholder

    @rank_zero_only
    def create_output_path(self) -> None:
        self.dirpath = Path(self.config.system.output.profiler)
        self.dirpath.mkdir(parents=True, exist_ok=True)

    def broadcast_profiler_path(self, string_var: str, src_rank: int) -> str:
        from lightning_fabric.utilities.distributed import group as _group

        string_var = [string_var]
        dist.broadcast_object_list(string_var, src_rank, group=_group.WORLD)
        return string_var[0]

    def setup(self, stage: str, local_rank: int | None = None, log_dir: str | None = None) -> None:
        del log_dir
        # THE STRATEGY IS ALREADY INITIALISED AND TORCH DISTRIBUTED IS ACTIVE
        # we need to broadcast the profiler path to all ranks to save the memory traces
        self.dirpath = Path(self.broadcast_profiler_path(str(self.dirpath), 0))
        self._stage = stage
        self._local_rank = local_rank
        self._create_time_profilers()
        self._create_memory_profilers()

    def _create_time_profilers(self) -> None:
        """Creates profilers for time and memory measurements."""
        if self.config.diagnostics.benchmark_profiler.time.enabled:
            self.time_profiler = SimpleProfiler(
                dirpath=self.dirpath,
            )

    def _create_memory_profilers(self) -> None:
        if self.config.diagnostics.benchmark_profiler.memory.enabled:
            import os

            def trace_handler(dir_name: str, stage: str | None = None) -> callable:

                def handler_fn(prof: pl.profilers.Profiler) -> None:
                    import socket
                    import time

                    worker_name = f"{socket.gethostname()}_{os.getpid()}"
                    file_name = str(dir_name / f"{worker_name}.{stage}.{time.time_ns()}.pt.trace.json")
                    LOGGER.info("Saving memory trace to %s", file_name)
                    prof.export_chrome_trace(file_name)

                return handler_fn

            global_rank = int(os.environ.get("SLURM_PROCID", "0"))  # WON'T WORK WHEN RUNNING WITHOUT SLURM
            if not (self.config.diagnostics.benchmark_profiler.memory.trace_rank0_only and global_rank != 0):
                from pytorch_lightning.profilers.pytorch import _KINETO_AVAILABLE

                assert (
                    _KINETO_AVAILABLE
                ), "Kineto is not available. Please ensure Kineto is avaialble to be able to use the memory profiler"

                torch.profiler.profile = (
                    PatchedProfile  # patch the profile(KinetoProfile) object to serialise the distributed info
                )
                self.memory_profiler = PyTorchProfiler(
                    with_stack=True,
                    emit_nvtx=False,
                    profile_memory=True,
                    export_to_chrome=True,
                    record_shapes=True,
                    group_by_input_shapes=True,
                    dirpath=self.dirpath,
                    on_trace_ready=trace_handler(self.dirpath),
                    schedule=torch.profiler.schedule(
                        wait=0,
                        warmup=self.warmup,
                        active=self.num_steps,
                        repeat=1,
                        skip_first=self.config.training.num_sanity_val_steps,
                    ),
                )
        self.time_rows_dict = None  # updated if we create a memory profile report

    def start(self, action_name: str) -> None:
        """Starts recording for a specific action.

        Parameters
        ----------
        action_name : str
            Name of the action.
        """
        self.time_profiler.start(action_name)
        self.memory_profiler.start(action_name)

    def stop(self, action_name: str) -> None:
        """Stops recording for a specific action.

        Parameters
        ----------
        action_name : str
            Name of the action.
        """
        self.time_profiler.stop(action_name)
        self.memory_profiler.stop(action_name)

    def _trim_time_report(self, recorded_actions: dict) -> dict[str, float]:
        all_actions_names = recorded_actions.keys()
        df = pd.DataFrame({"Strings": all_actions_names})
        combined_pattern = "|".join(PROFILER_ACTIONS)
        filtered_df = df[df["Strings"].str.contains(combined_pattern, regex=True, na=False)]
        trimmed_actions_names = filtered_df["Strings"].tolist()
        return {key: recorded_actions[key] for key in trimmed_actions_names}

    def get_time_profiler_df(self, precision: int = 5) -> pd.DataFrame:
        """Retrieves a DataFrame with time profiling information.

        Parameters
        ----------
        precision : int
            Precision for rounding, by default 5

        Returns
        -------
        pd.DataFrame
            DataFrame with time profiling information.
        """
        if self.config.diagnostics.benchmark_profiler.time.verbose is False:
            self.time_profiler.recorded_durations = self._trim_time_report(
                recorded_actions=self.time_profiler.recorded_durations,
            )
        time_df = pd.DataFrame(self.time_profiler.recorded_durations.items())
        time_df[2] = time_df[1].apply(len)
        time_df[3] = time_df[1].apply(np.mean)
        time_df[1] = time_df[1].apply(sum)
        time_df.columns = ["name", "total_time", "n_calls", "avg_time"]

        def replace_function(value: str) -> str:
            # Replace 'apple' with 'fruit'
            return re.sub(r"\{.*?\}", "", value)  # Remove anything between braces

        time_df["name"] = time_df["name"].apply(replace_function)
        pattern = r"\[(.*?)\]|(.*)"
        time_df["category"] = time_df["name"].str.extract(pattern, expand=False)[0].fillna(time_df["name"])

        pattern = re.compile(r"\[Callback\](.*?)\.")
        # Apply the regular expression to the column
        callbacks_subcategories = "*Callback_" + time_df[time_df["category"] == "Callback"]["name"].str.extract(pattern)
        indexer = time_df[time_df["category"] == "Callback"].index
        time_df.loc[indexer, "category"] = callbacks_subcategories[0].tolist()

        # Check if 'Callback' is present in the 'category' column
        time_df["is_callback"] = time_df["category"].str.contains("Callback", case=False)

        # Group by the 'is_callback' column and apply groupby operation only on rows with 'Callback' in 'category'
        grouped_data = (
            time_df[time_df["is_callback"]]
            .groupby("category")
            .agg({"n_calls": "sum", "avg_time": "sum", "total_time": "sum"})
            .reset_index()
        )
        grouped_data["name"] = grouped_data["category"]

        time_df = pd.concat([time_df[~time_df["is_callback"]], grouped_data])
        time_df = time_df.drop("is_callback", axis=1)
        time_df = time_df.round(precision)
        time_df = time_df.sort_values(by="category", ascending=False)

        self.time_report_fname = self.dirpath / "time_profiler.csv"
        self._save_report(time_df, self.time_report_fname)
        return time_df

    @staticmethod
    def to_df(sample_dict: dict[str, float], precision: str = ".5") -> pd.DataFrame:
        df = pd.DataFrame(sample_dict.items())
        df.columns = ["metric", "value"]
        df.value = df.value.apply(lambda x: f"%{precision}f" % x)
        return df

    @rank_zero_only
    def get_system_profiler_df(self, logger_name: str, logger: pl.loggers.Logger) -> pd.DataFrame:
        if logger_name == "wandb":
            system_metrics_df = self.to_df(WandBSystemSummarizer(logger).summarize_system_metrics())
        elif logger_name == "mlflow":
            system_metrics_df = MLFlowSystemSummarizer(logger).summarize_mlflow_system_metrics()
        elif logger_name == "tensorboard":
            LOGGER.info("No system profiler data available for Tensorboard")
            system_metrics_df = None

        self.system_report_fname = self.dirpath / "system_profiler.csv"
        self._save_report(system_metrics_df, self.system_report_fname)
        return system_metrics_df

    def _save_report(self, df: pd.DataFrame, fname: Path) -> None:
        df.to_csv(fname)

    def _save_model_summary(self, model_summary: str, fname: Path) -> None:
        with fname.open("w") as f:
            f.write(model_summary)
            f.close()

    def get_model_summary(self, model: BaseGraphModule, example_input_array: np.ndarray) -> str:

        from torchinfo import summary

        # when using flash attention model, we need to convert the input and model to float16 and cuda
        # since FlashAttention only supports fp16 and bf16 data type
        example_input_array = example_input_array.to(dtype=torch.float16)
        example_input_array = example_input_array.to("cuda")
        model.half()
        model = model.to("cuda")

        summary_str = str(
            summary(
                model,
                input_data=example_input_array,
                depth=20,
                col_width=16,
                col_names=["trainable", "input_size", "output_size", "num_params", "params_percent", "mult_adds"],
                row_settings=["var_names"],
                verbose=0,
            ),
        )
        self.model_summary_fname = self.dirpath / "model_summary.txt"
        self._save_model_summary(summary_str, self.model_summary_fname)
        return summary_str

    @rank_zero_only
    def get_speed_profiler_df(self, progressbar: _tqdm) -> pd.DataFrame:
        """Computes the speed metrics based on training and validation rates."""
        speed_metrics = {}

        batch_size_tr = self.config.dataloader.batch_size.training
        batch_size_val = self.config.dataloader.batch_size.validation

        training_rates_array = np.array(progressbar.training_rates)
        speed_metrics["training_avg_throughput"] = training_rates_array.mean()
        speed_metrics["training_avg_throughput_per_sample"] = training_rates_array.mean() / batch_size_tr

        validation_rates_array = np.array(progressbar.validation_rates)
        speed_metrics["validation_avg_throughput"] = validation_rates_array.mean()
        speed_metrics["validation_avg_throughput_per_sample"] = validation_rates_array.mean() / batch_size_val

        # Calculate per_sample metrics
        speed_metrics["avg_training_dataloader_throughput"] = (
            1 / np.array(self.time_profiler.recorded_durations["[_TrainingEpochLoop].train_dataloader_next"]).mean()
        )
        speed_metrics["avg_training_dataloader_throughput_per_sample"] = (
            speed_metrics["avg_training_dataloader_throughput"] / batch_size_tr
        )

        speed_metrics["avg_validation_dataloader_throughput"] = (
            1 / np.array(self.time_profiler.recorded_durations["[_EvaluationLoop].val_next"]).mean()
        )
        speed_metrics["avg_validation_dataloader_throughput_per_sample"] = (
            speed_metrics["avg_validation_dataloader_throughput"] / batch_size_val
        )

        if self.time_rows_dict:
            speed_metrics.update(self.time_rows_dict)

        speed_profile_df = self.to_df(speed_metrics)

        self.speed_report_fname = self.dirpath / "speed_profiler.csv"
        self._save_report(speed_profile_df, self.speed_report_fname)

        return speed_profile_df

    def _save_extra_plots(self) -> None:
        if check_torch_version():
            # !it's available for  torch >= 2.1
            from torch.cuda._memory_viz import profile_plot

            self.memory_trace_fname = Path(self.dirpath, "memory_trace.html")
            with self.memory_trace_fname.open("w") as f:
                f.write(profile_plot(self.memory_profiler.profiler))

            # !it's available for  torch >= 2.1
            self.memory_timeline_fname = str(Path(self.dirpath, "memory_timelines.html"))
            self.memory_profiler.profiler.export_memory_timeline(self.memory_timeline_fname)

    @rank_zero_only
    def get_memory_profiler_df(self) -> pd.DataFrame:
        """Retrieves the memory profiler data as a DataFrame.

        Aggregates the results coming from multiple nodes/processes.

        Returns
        -------
        pd.DataFrame
            Memory profiler data.
        """
        if self.config.diagnostics.benchmark_profiler.memory.extra_plots:
            self._save_extra_plots()

        self.memory_profiler._delete_profilers()

        if not self.memory_profiler.function_events:
            return ""

        data = self.memory_profiler.function_events.key_averages(
            group_by_input_shapes=self.memory_profiler._group_by_input_shapes,
        )
        table = data.table(
            sort_by=self.memory_profiler._sort_by_key,
            row_limit=self.memory_profiler._row_limit,
            **self.memory_profiler._table_kwargs,
        )  # this is a string

        from io import StringIO

        table_main_body = table.split("\n")[:-3]  # Remove the last rows
        columns = [
            "Name",
            "Self CPU %",
            "Self CPU",
            "CPU total %",
            "CPU total",
            "CPU time avg",
            "Self CUDA",
            "Self CUDA %",
            "CUDA total",
            "CUDA time avg",
            "CPU Mem",
            "Self CPU Mem",
            "CUDA Mem",
            "Self CUDA Mem",
            "# of Calls",
            "Input Shapes",
        ]
        table_main_body = "\n".join(table_main_body)
        memory_df = pd.read_fwf(StringIO(table_main_body), names=columns, skiprows=2)
        flag = ["--" not in row for row in memory_df["Name"]]
        memory_df = memory_df[flag]
        time_rows = [row for row in table.split("\n")[-3:] if row != ""]
        try:
            if time_rows:
                time_rows_dict = {}
                for row in time_rows:
                    key, val = row.split(":")
                    val = convert_to_seconds(val.strip())
                    time_rows_dict[key] = val
                self.time_rows_dict = time_rows_dict

                memory_df = memory_df[~memory_df["Name"].isin(time_rows)]
        except ValueError as e:
            LOGGER.info("Error saving memory df: %s", e)

        self.memory_report_fname = self.dirpath / "memory_profiler.csv"
        self._save_report(memory_df, self.memory_report_fname)
        return memory_df


class ProfilerProgressBar(TQDMProgressBar):
    """Custom PyTorch Lightning progress bar with profiling functionality.

    Attributes
    ----------
    validation_rates : list[float]
        List to store validation rates (it/s).
    training_rates : list[float]
        List to store training rates (it/s).
    """

    def __init__(self, refresh_rate: int = 1):
        super().__init__()
        self.validation_rates = []
        self.training_rates = []
        self._refresh_rate = refresh_rate

    def _extract_rate(self, pbar: _tqdm) -> float:
        """Extracts the iteration rate from the progress bar.

        Parameters
        ----------
        pbar : tqdm
            The progress bar.

        Returns
        -------
        float
            The iteration rate.
        """
        return (pbar.format_dict["n"] - pbar.format_dict["initial"]) / pbar.format_dict["elapsed"]

    def on_train_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
    ) -> None:
        """Appends the rate from the progress bar to the list of 'training_rates'."""
        super().on_train_batch_end(trainer, pl_module, outputs, batch, batch_idx)
        if self.train_progress_bar.format_dict["n"] != 0:
            self.training_rates.append(self._extract_rate(self.train_progress_bar))
            for logger in self.trainer.loggers:
                if isinstance(logger, AnemoiMLflowLogger):
                    logger.log_metrics({"training_rate": self.training_rates[-1]}, step=trainer.global_step)

    def on_validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Append rate from the progress bar to the list of 'validation_rates'."""
        super().on_validation_batch_end(trainer, pl_module, outputs, batch, batch_idx, dataloader_idx)
        if self.val_progress_bar.format_dict["n"] != 0:
            self.validation_rates.append(self._extract_rate(self.val_progress_bar))
            for logger in self.trainer.loggers:
                if isinstance(logger, AnemoiMLflowLogger):
                    logger.log_metrics({"validation_rate": self.validation_rates[-1]}, step=trainer.global_step)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
from pathlib import Path

import torch
from omegaconf import DictConfig
from omegaconf import ListConfig
from omegaconf import OmegaConf

from anemoi.models.data_indices.collection import BaseIndex
from anemoi.models.data_indices.collection import IndexCollection
from anemoi.models.data_indices.tensor import BaseTensorIndex
from anemoi.utils.dates import frequency_to_string


def map_config_to_primitives(config: OmegaConf) -> dict:
    """Ensure that the metadata information is JSON-serializable.

    Parameters
    ----------
    config : OmegaConf
        config object to be mapped to primitives.

    Returns
    -------
    dict
        JSON serializable dictionary.

    Raises
    ------
    TypeError
        Cannot map object to primitives.

    """
    if config is None or isinstance(config, int | float | str | bool):
        return config

    if isinstance(config, Path):
        config = str(config)
    elif isinstance(config, datetime.date):
        config = config.isoformat()
    elif isinstance(config, datetime.timedelta):
        config = frequency_to_string(config)
    elif isinstance(config, list | tuple):
        config = [map_config_to_primitives(v) for v in config]
    elif isinstance(config, dict):
        config = {k: map_config_to_primitives(v) for k, v in config.items()}
    elif isinstance(config, DictConfig | ListConfig):
        config = map_config_to_primitives(OmegaConf.to_container(config, resolve=True))
    elif isinstance(config, torch.Tensor):
        config = map_config_to_primitives(config.tolist())
    elif isinstance(config, IndexCollection | BaseTensorIndex | BaseIndex):
        config = map_config_to_primitives(config.todict())
    else:
        msg = f"Cannot serialize object of type {type(config)}"
        raise TypeError(msg)

    return config



from typing import TYPE_CHECKING

import pytest

from anemoi.utils.testing import cli_testing

if TYPE_CHECKING:
    from pytest_mock import MockerFixture
    from pytest_mock import MockType


@pytest.fixture
def mock_auth(mocker: "MockerFixture") -> "MockType":
    mock_auth = mocker.patch("anemoi.utils.mlflow.auth.TokenAuth")
    mock_auth.get_servers.return_value = [("http://server-2", 2), ("http://server-1", 1)]
    return mock_auth


def test_mlflow_login(mocker: "MockerFixture", mock_auth: "MockType") -> None:
    cli_testing("anemoi-training", "mlflow", "login", "--url", "http://localhost:5000")
    mock_auth.assert_called_once_with(url="http://localhost:5000")
    mock_auth.return_value.login.assert_called_once()
    mock_auth.reset_mock()

    cli_testing("anemoi-training", "mlflow", "login")
    mock_auth.get_servers.assert_called_once()
    mock_auth.assert_called_once_with(url="http://server-2")
    mock_auth.return_value.login.assert_called_once()
    mock_auth.reset_mock()

    cli_testing("anemoi-training", "mlflow", "login", "--list")
    mock_auth.get_servers.assert_called_once()
    mock_auth.return_value.login.assert_not_called()
    mock_auth.reset_mock()

    cli_testing("anemoi-training", "mlflow", "login", "--all")
    mock_auth.get_servers.assert_called_once()
    assert mock_auth.call_args_list == [mocker.call(url="http://server-2"), mocker.call(url="http://server-1")]
    assert mock_auth.return_value.login.call_count == 2
    mock_auth.reset_mock()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import torch
from omegaconf import DictConfig

from anemoi.models.data_indices.collection import IndexCollection
from anemoi.training.losses.scalers.base_scaler import BaseScaler
from anemoi.training.utils.enums import TensorDim
from anemoi.training.utils.variables_metadata import ExtractVariableGroupAndLevel

LOGGER = logging.getLogger(__name__)


class BaseVariableLossScaler(BaseScaler):
    """Base class for all variable loss scalers."""

    scale_dims: TensorDim = TensorDim.VARIABLE

    def __init__(
        self,
        data_indices: IndexCollection,
        metadata_extractor: ExtractVariableGroupAndLevel,
        norm: str | None = None,
        **kwargs,
    ) -> None:
        """Initialise Scaler.

        Parameters
        ----------
        data_indices : IndexCollection
            Collection of data indices.
        metadata_extractor : ExtractVariableGroupAndLevel
            Metadata extractor for variable groups and levels.
        norm : str, optional
            Type of normalization to apply. Options are None, unit-sum, unit-mean and l1.
        """
        super().__init__(norm=norm)
        del kwargs
        self.data_indices = data_indices
        self.variable_metadata_extractor = metadata_extractor


class GeneralVariableLossScaler(BaseVariableLossScaler):
    """Scaling per variable defined in config file."""

    def __init__(
        self,
        data_indices: IndexCollection,
        weights: DictConfig,
        metadata_extractor: ExtractVariableGroupAndLevel,
        norm: str | None = None,
        **kwargs,
    ) -> None:
        """Initialise GeneralVariableLossScaler.

        Parameters
        ----------
        data_indices : IndexCollection
            Collection of data indices.
        weights : DictConfig
            Configuration for variable loss scaling.
        scale_dim : int
            Dimension to scale
        metadata_extractor : ExtractVariableGroupAndLevel
            Metadata extractor for variable groups and levels.
        norm : str, optional
            Type of normalization to apply. Options are None, unit-sum, unit-mean and l1.
        """
        super().__init__(data_indices, metadata_extractor=metadata_extractor, norm=norm)
        self.weights = weights
        del kwargs

    def get_scaling_values(self, **_kwargs) -> torch.Tensor:
        """Get loss scaling.

        Retrieve the loss scaling for each variable from the config file.
        """
        variable_loss_scaling = torch.empty((len(self.data_indices.data.output.full),), dtype=torch.float32)

        for variable_name, idx in self.data_indices.model.output.name_to_index.items():
            _, variable_ref, _ = self.variable_metadata_extractor.get_group_and_level(variable_name)
            # Apply variable scaling by variable name
            # or base variable name (variable_ref: variable name without variable level)
            variable_loss_scaling[idx] = self.weights.get(
                variable_ref,
                self.weights.get("default", 1.0),
            )
            if variable_ref != variable_name:
                assert (
                    self.weights.get(
                        variable_name,
                        None,
                    )
                    is None
                ), f"Variable {variable_name} is not allowed to have a separate scaling besides {variable_ref}."

        return variable_loss_scaling



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import numpy as np

from anemoi.training.utils.usable_indices import get_usable_indices


def test_get_usable_indices() -> None:
    """Test get_usable_indices function."""
    # Test base case
    valid_indices = get_usable_indices(missing_indices=set(), series_length=10, relative_indices=np.array([0, 1]))
    expected_values = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
    assert np.allclose(valid_indices, expected_values)

    # Test 3 indices, either from rollout = 1 and multistep = 2, or rollout = 2 and multistep = 1
    valid_indices = get_usable_indices(missing_indices=set(), series_length=10, relative_indices=np.array([0, 1, 2]))
    expected_values = np.array([0, 1, 2, 3, 4, 5, 6, 7])
    assert np.allclose(valid_indices, expected_values)

    # With time increment
    valid_indices = get_usable_indices(missing_indices=set(), series_length=10, relative_indices=np.array([0, 2, 4]))
    expected_values = np.array([0, 1, 2, 3, 4, 5])
    assert np.allclose(valid_indices, expected_values)

    # Test missing indices with standard setup
    missing_indices = {7, 5, 14}
    valid_indices = get_usable_indices(
        missing_indices=missing_indices,
        series_length=20,
        relative_indices=np.array([0, 1, 2]),
    )
    expected_values = np.array([0, 1, 2, 8, 9, 10, 11, 15, 16, 17])
    assert np.allclose(valid_indices, expected_values)

    # Now verify that with a non-standard setup, missing indices can be "jumped" over.
    valid_indices = get_usable_indices(
        missing_indices=missing_indices,
        series_length=20,
        relative_indices=np.array([0, 5, 6, 7]),  # e.g making a 1 hour forecast based on -6, -1 and 0 hours.
    )
    expected_values = np.array([3, 4, 6, 10, 11, 12])
    assert np.allclose(valid_indices, expected_values)

    # Test functionality for avoiding using different model trajectories
    series_length = 70
    mr_start, mr_length = [4, 18]
    trajectory_ids = (np.arange(series_length, dtype=np.int64) - mr_start) // mr_length
    valid_indices = get_usable_indices(
        missing_indices=set(range(40, 58)) | {11},  # one model run of length 18 missing, and one sample of another run.
        series_length=series_length,
        relative_indices=np.array([0, 3, 6]),
        trajectory_ids=trajectory_ids,
    )
    expected_values = np.array([*[4, 6, 7, 9, 10, 12, 13, 14, 15], *list(range(22, 40 - 6)), *[58, 59, 60, 61, 62, 63]])



# (C) Copyright 2024- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
from typing import Annotated
from typing import Any
from typing import Literal

from omegaconf import OmegaConf
from pydantic import Field
from pydantic import NonNegativeInt
from pydantic import PositiveInt
from pydantic import model_validator
from pydantic import root_validator

from anemoi.training.diagnostics.mlflow import MAX_PARAMS_LENGTH
from anemoi.utils.schemas import BaseModel

LOGGER = logging.getLogger(__name__)


class GraphTrainableFeaturesPlotSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.callbacks.plot.GraphTrainableFeaturesPlot"] = Field(alias="_target_")
    "GraphTrainableFeaturesPlot object from anemoi training diagnostics callbacks."
    every_n_epochs: int | None
    "Epoch frequency to plot at."


class PlotLossSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.callbacks.plot.PlotLoss"] = Field(alias="_target_")
    "PlotLoss object from anemoi training diagnostics callbacks."
    parameter_groups: dict[str, list[str]]
    "Dictionary with parameter groups with parameter names as key."
    every_n_batches: int | None = Field(default=None)
    "Batch frequency to plot at."


class MatplotlibColormapSchema(BaseModel):
    target_: Literal["anemoi.training.utils.custom_colormaps.MatplotlibColormap"] = Field(..., alias="_target_")
    "CustomColormap object from anemoi training utils."
    name: str
    "Name of the Matplotlib colormap."
    variables: list[str] | None = Field(default=None)
    "A list of strings representing the variables for which the colormap is used, by default None."


class MatplotlibColormapClevelsSchema(BaseModel):
    target_: Literal["anemoi.training.utils.custom_colormaps.MatplotlibColormapClevels"] = Field(..., alias="_target_")
    "CustomColormap object from anemoi training utils."
    clevels: list
    "The custom color levels for the colormap."
    variables: list[str] | None = Field(default=None)
    "A list of strings representing the variables for which the colormap is used, by default None."


class DistinctipyColormapSchema(BaseModel):
    target_: Literal["anemoi.training.utils.custom_colormaps.DistinctipyColormap"] = Field(..., alias="_target_")
    "CustomColormap object from anemoi training utils."
    n_colors: int
    "The number of colors in the colormap."
    variables: list[str] | None = Field(default=None)
    "A list of strings representing the variables for which the colormap is used, by default None."
    colorblind_type: str | None = Field(default=None)
    "The type of colorblindness to simulate. If None, the default colorblindness from distinctipy is applied."


ColormapSchema = Annotated[
    MatplotlibColormapSchema | MatplotlibColormapClevelsSchema | DistinctipyColormapSchema,
    Field(discriminator="target_"),
]


class LongRolloutPlotsSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.callbacks.plot.LongRolloutPlots"] = Field(alias="_target_")
    "LongRolloutPlots object from anemoi training diagnostics callbacks."
    rollout: list[int]
    "Rollout steps to plot at."
    sample_idx: int
    "Index of sample to plot, must be inside batch size."
    parameters: list[str]
    "List of parameters to plot."
    video_rollout: int = Field(example=0)
    "Number of rollout steps for video, by default 0 (no video)."
    accumulation_levels_plot: list[float] | None = Field(default=None)
    "Accumulation levels to plot, by default None."
    cmap_accumulation: list[str] | None = Field(default=None)
    "Colors of the accumulation levels. Default to None. Kept for backward compatibility."
    per_sample: int | None = Field(default=None)
    "Number of plots per sample, by default 6."
    every_n_epochs: int = Field(example=1)
    "Epoch frequency to plot at, by default 1."
    animation_interval: int | None = Field(default=None)
    "Delay between frames in the animation in milliseconds, by default 400."
    colormaps: dict[str, ColormapSchema] | None = Field(default=None)
    "List of colormaps to use, by default None."


class PlotSampleSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.callbacks.plot.PlotSample"] = Field(alias="_target_")
    "PlotSample object from anemoi training diagnostics callbacks."
    sample_idx: int
    "Index of sample to plot, must be inside batch size."
    parameters: list[str]
    "List of parameters to plot."
    accumulation_levels_plot: list[float]
    "Accumulation levels to plot."
    cmap_accumulation: list[str] | None = Field(default=None)
    "Colors of the accumulation levels. Default to None. Kept for backward compatibility."
    precip_and_related_fields: list[str] | None = Field(default=None)
    "List of precipitation related fields, by default None."
    per_sample: int = Field(example=6)
    "Number of plots per sample, by default 6."
    every_n_batches: int | None = Field(default=None)
    "Batch frequency to plot at, by default None."
    colormaps: dict[str, ColormapSchema] | None = Field(default=None)
    "List of colormaps to use, by default None."


class PlotSpectrumSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.callbacks.plot.PlotSpectrum"] = Field(alias="_target_")
    "PlotSpectrum object from anemoi training diagnostics callbacks."
    sample_idx: int
    "Index of sample to plot, must be inside batch size."
    parameters: list[str]
    "List of parameters to plot."
    every_n_batches: int | None = Field(default=None)
    "Batch frequency to plot at, by default None."


class PlotHistogramSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.callbacks.plot.PlotHistogram"] = Field(alias="_target_")
    "PlotHistogram object from anemoi training diagnostics callbacks."
    sample_idx: int
    "Index of sample to plot, must be inside batch size."
    parameters: list[str]
    "List of parameters to plot."
    precip_and_related_fields: list[str] | None = Field(default=None)
    "List of precipitation related fields, by default None."
    every_n_batches: int | None = Field(default=None)
    "Batch frequency to plot at, by default None."


class PlotEnsSampleSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.callbacks.plot_ens.PlotEnsSample"] = Field(alias="_target_")
    "PlotEnsSample object from anemoi training diagnostics callbacks."
    sample_idx: int
    "Index of sample to plot, must be inside batch size."
    parameters: list[str]
    "List of parameters to plot."
    accumulation_levels_plot: list[float]
    "Accumulation levels to plot."
    cmap_accumulation: list[str] | None = Field(default=None)
    "Colors of the accumulation levels. Default to None. Kept for backward compatibility."
    precip_and_related_fields: list[str] | None = Field(default=None)
    "List of precipitation related fields, by default None."
    per_sample: int = Field(example=6)
    "Number of plots per sample, by default 6."
    every_n_batches: int | None = Field(default=None)
    "Batch frequency to plot at, by default None."
    colormaps: dict[str, ColormapSchema] | None = Field(default=None)
    "List of colormaps to use, by default None."
    members: list[int] | None = Field(default=None)
    "List of ensemble members to plot. If None, plots all members."


class PlotEnsLossSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.callbacks.plot_ens.PlotLoss"] = Field(alias="_target_")
    "PlotLoss object from anemoi training diagnostics callbacks."
    parameter_groups: dict[str, list[str]]
    "Dictionary with parameter groups with parameter names as key."
    every_n_batches: int | None = Field(default=None)
    "Batch frequency to plot at."


class PlotEnsSpectrumSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.callbacks.plot_ens.PlotSpectrum"] = Field(alias="_target_")
    "PlotSpectrum object from anemoi training diagnostics callbacks."
    sample_idx: int
    "Index of sample to plot, must be inside batch size."
    parameters: list[str]
    "List of parameters to plot."
    every_n_batches: int | None = Field(default=None)
    "Batch frequency to plot at, by default None."


class PlotEnsHistogramSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.callbacks.plot_ens.PlotHistogram"] = Field(alias="_target_")
    "PlotHistogram object from anemoi training diagnostics callbacks."
    sample_idx: int
    "Index of sample to plot, must be inside batch size."
    parameters: list[str]
    "List of parameters to plot."
    precip_and_related_fields: list[str] | None = Field(default=None)
    "List of precipitation related fields, by default None."
    every_n_batches: int | None = Field(default=None)
    "Batch frequency to plot at, by default None."


class GraphTrainableFeaturesPlotEnsSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.callbacks.plot_ens.GraphTrainableFeaturesPlot"] = Field(
        alias="_target_",
    )
    "GraphTrainableFeaturesPlot object from anemoi training diagnostics callbacks."
    every_n_epochs: int | None
    "Epoch frequency to plot at."


PlotCallbacks = Annotated[
    LongRolloutPlotsSchema
    | GraphTrainableFeaturesPlotSchema
    | PlotLossSchema
    | PlotSampleSchema
    | PlotSpectrumSchema
    | PlotHistogramSchema
    | PlotEnsSampleSchema
    | PlotEnsLossSchema
    | PlotEnsSpectrumSchema
    | PlotEnsHistogramSchema
    | GraphTrainableFeaturesPlotEnsSchema,
    Field(discriminator="target_"),
]


class PlottingFrequency(BaseModel):
    batch: PositiveInt = Field(example=750)
    "Frequency of the plotting in number of batches."
    epoch: PositiveInt = Field(example=5)
    "Frequency of the plotting in number of epochs."


class PlotSchema(BaseModel):
    asynchronous: bool
    "Handle plotting tasks without blocking the model training."
    datashader: bool
    "Use Datashader to plot."
    frequency: PlottingFrequency
    "Frequency of the plotting."
    sample_idx: int
    "Index of sample to plot, must be inside batch size."
    parameters: list[str]
    "List of parameters to plot."
    precip_and_related_fields: list[str]
    "List of precipitation related fields from the parameters list."
    colormaps: dict[str, ColormapSchema] = Field(default_factory=dict)
    "List of colormaps to use."
    callbacks: list[PlotCallbacks] = Field(example=[])
    "List of plotting functions to call."


class TimeLimitSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.callbacks.stopping.TimeLimit"] = Field(alias="_target_")
    "TimeLimit object from anemoi training diagnostics callbacks."
    limit: int | str
    "Time limit, if int, assumed to be hours, otherwise must be a string with units (e.g. '1h', '30m')."
    record_file: str | None = Field(default=None)
    "File to record the last checkpoint to on exit, if set."


class EarlyStoppingSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.callbacks.stopping.EarlyStopping"] = Field(alias="_target_")
    monitor: str = Field(examples=["val_wmse_epoch", "val_wmse/sfc_2t/1"])
    "Metric to monitor"
    min_delta: float = 0.0
    "Minimum change in the monitored quantity to qualify as an improvement."
    patience: int = 3
    "Number of epochs with no improvement after which training will be stopped."
    verbose: bool = False
    "If True, prints a message for each improvement."
    mode: Literal["min", "max"] = "min"
    "One of {'min', 'max'}, changes if minimisation or maximimisation of the metric is 'good'."
    strict: bool = True
    "Whether to crash the training if the monitored quantity is not found."
    check_finite: bool = True
    "Whether to check for NaNs and Infs in the monitored quantity."
    stopping_threshold: float | None = None
    "Stop training immediately once the monitored quantity reaches this threshold."
    divergence_threshold: float | None = None
    "Stop training as soon as the monitored quantity becomes worse than this threshold.."
    check_on_train_epoch_end: bool | None = None
    "Whether to check the stopping criteria at the end of each training epoch."


class Debug(BaseModel):
    anomaly_detection: bool
    "Activate anomaly detection. This will detect and trace back NaNs/Infs, but slow down training."


class CheckpointSchema(BaseModel):
    save_frequency: int | None
    "Frequency at which to save the checkpoints."
    num_models_saved: int
    "Number of model checkpoint to save. Only the last num_models_saved checkpoints will be kept. \
            If set to -1, all checkpoints are kept"


class WandbSchema(BaseModel):
    enabled: bool
    "Use Weights & Biases logger."
    offline: bool
    "Run W&B offline."
    log_model: bool | Literal["all"]
    "Log checkpoints created by ModelCheckpoint as W&B artifacts. \
            If True, checkpoints are logged at the end of training. If 'all', checkpoints are logged during training."
    project: str
    "The name of the project to which this run will belong."
    gradients: bool
    "Whether to log the gradients."
    parameters: bool
    "Whether to log the hyper parameters."
    entity: str | None = None
    "Username or team name where to send runs. This entity must exist before you can send runs there."

    @root_validator(pre=True)
    def clean_entity(cls: type["WandbSchema"], values: dict[str, Any]) -> dict[str, Any]:  # noqa: N805
        if values["enabled"] is False:
            values["entity"] = None
        return values


class MlflowSchema(BaseModel):
    target_: Literal["anemoi.training.diagnostics.mlflow.logger.AnemoiMLflowLogger"] = Field(
        default="anemoi.training.diagnostics.mlflow.logger.AnemoiMLflowLogger",
        alias="_target_",
    )
    enabled: bool
    "Use MLflow logger."
    offline: bool
    "Run MLflow offline. Necessary if no internet access available."
    authentication: bool
    "Whether to authenticate with server or not"
    log_model: bool | Literal["all"] | None = None
    "Log checkpoints created by ModelCheckpoint as MLFlow artifacts. \
            If True, checkpoints are logged at the end of training. If 'all', checkpoints are logged during training."
    tracking_uri: str | None = None
    "Address of local or remote tracking server."
    experiment_name: str
    "Name of experiment."
    project_name: str
    "Name of project."
    system: bool
    "Activate system metrics."
    terminal: bool
    "Log terminal logs to MLflow."
    run_name: str | None
    "Name of run."
    on_resume_create_child: bool
    "Whether to create a child run when resuming a run."
    expand_hyperparams: list[str] = Field(default_factory=lambda: ["config"])
    "Keys to expand within params. Any key being expanded will have lists converted according to `expand_iterables`."
    http_max_retries: PositiveInt = Field(example=35)
    "Specifies the maximum number of retries for MLflow HTTP requests, default 35."
    max_params_length: int = MAX_PARAMS_LENGTH
    "Maximum number of hpParams to be logged with mlflow"
    save_dir: str | None = None
    "Directory to save logs to when offline=True, default={system.output.root}/{system.output.logs.mlflow}"

    @root_validator(pre=True)
    def clean_entity(cls: type["MlflowSchema"], values: dict[str, Any]) -> dict[str, Any]:  # noqa: N805
        if values["enabled"] is False:
            values["tracking_uri"] = None
        return values


class AzureMlflowSchema(MlflowSchema):
    target_: Literal["anemoi.training.diagnostics.mlflow.azureml.AnemoiAzureMLflowLogger"] = Field(
        ...,
        alias="_target_",
    )

    # These options are inherited, but either don't't make sense or don't work for Azure
    # so we enforce the required value
    offline: Literal[False]
    terminal: Literal[False]
    # These are specific to Azure
    identity: str | None = None
    "Type of identity to use for logging in with Azure ML."
    resource_group: str | None = None
    "Name of the AzureML resource group"
    workspace_name: str | None = None
    "Name of the AzureML workspace"
    subscription_id: str | None = None
    "AzureML subscription ID"
    azure_log_level: str = "WARNING"
    "Log level for all azure packages (azure-identity, azure-core, etc)"


class TensorboardSchema(BaseModel):
    enabled: bool
    "Use TensorBoard logger."


class LoggingSchema(BaseModel):
    wandb: WandbSchema
    "W&B logging schema."
    tensorboard: TensorboardSchema
    "TensorBorad logging schema."
    mlflow: Annotated[MlflowSchema | AzureMlflowSchema, Field(discriminator="target_")]
    "MLflow logging schema."
    interval: PositiveInt
    "Logging frequency in batches."

    @model_validator(mode="before")
    def inject_default_target(cls: type["MlflowSchema"], values: dict[str, Any]) -> dict[str, Any]:  # noqa: N805
        mlflow_cfg = OmegaConf.to_container(values.get("mlflow"), resolve=True)
        if mlflow_cfg is not None and isinstance(mlflow_cfg, dict) and "_target_" not in mlflow_cfg:
            mlflow_cfg["_target_"] = "anemoi.training.diagnostics.mlflow.logger.AnemoiMLflowLogger"
            values["mlflow"] = mlflow_cfg
        return values


class MemorySchema(BaseModel):
    enabled: bool = Field(example=False)
    "Enable memory report. Default to false."
    steps: PositiveInt = Field(example=5)
    "Frequency of memory profiling. Default to 5."
    warmup: NonNegativeInt = Field(example=2)
    "Number of step to discard before the profiler starts to record traces. Default to 2."
    extra_plots: bool = Field(example=False)
    "Save plots produced with torch.cuda._memory_viz.profile_plot if available. Default to false."
    trace_rank0_only: bool = Field(example=False)
    "Trace only rank 0 from SLURM_PROC_ID. Default to false."


class Snapshot(BaseModel):
    enabled: bool = Field(example=False)
    "Enable memory snapshot recording. Default to false."
    steps: PositiveInt = Field(example=4)
    "Frequency of snapshot. Default to 4."
    warmup: NonNegativeInt = Field(example=0)
    "Number of step to discard before the profiler starts to record traces. Default to 0."


class Profiling(BaseModel):
    enabled: bool = Field(example=False)
    "Enable component profiler. Default to false."
    verbose: bool | None = None
    "Set to true to include the full list of profiled action or false to keep it concise."


class BenchmarkProfilerSchema(BaseModel):
    memory: MemorySchema = Field(default_factory=lambda: MemorySchema())
    "Schema for memory report containing metrics associated with CPU and GPU memory allocation."
    time: Profiling = Field(default_factory=lambda: Profiling(True))
    "Report with metrics of execution time for certain steps across the code."
    speed: Profiling = Field(default_factory=lambda: Profiling(True))
    "Report with metrics of execution speed at training and validation time."
    system: Profiling = Field(default_factory=lambda: Profiling())
    "Report with metrics of GPU/CPU usage, memory and disk usage and total execution time."
    model_summary: Profiling = Field(default_factory=lambda: Profiling())
    "Table summary of layers and parameters of the model."
    snapshot: Snapshot = Field(default_factory=lambda: Snapshot())
    "Memory snapshot if torch.cuda._record_memory_history is available."


class ProgressBarSchema(BaseModel):
    target_: Literal[
        "pytorch_lightning.callbacks.TQDMProgressBar",
        "pytorch_lightning.callbacks.RichProgressBar",
        "anemoi.training.diagnostics.profilers.ProfilerProgressBar",
    ] = Field(alias="_target_")
    "TQDMProgressBar object from pytorch lightning."
    refresh_rate: PositiveInt = Field(default=1)
    "Refresh rate of the progress bar."


class DiagnosticsSchema(BaseModel):
    plot: PlotSchema | None = None
    "Plot schema."
    callbacks: list = Field(default_factory=list, example=[])
    "Callbacks schema."
    benchmark_profiler: BenchmarkProfilerSchema
    "Benchmark profiler schema for `profile` command."
    debug: Debug
    "Debug schema."
    log: LoggingSchema
    "Log schema."
    enable_progress_bar: bool
    "Activate progress bar."
    progress_bar: ProgressBarSchema | None = Field(default=None)
    "Progress bar schema."
    print_memory_summary: bool
    "Print the memory summary."
    enable_checkpointing: bool
    "Allow model to save checkpoints."
    checkpoint: dict[str, CheckpointSchema] = Field(default_factory=dict)
    "Checkpoint schema for defined frequency (every_n_minutes, every_n_epochs, ...)."
    check_val_every_n_epoch: PositiveInt = Field(default=1, example=1)
    "Run validation every n epochs."



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from hydra import initialize
from hydra.core.global_hydra import GlobalHydra
from hydra.core.plugins import Plugins
from hydra.plugins.search_path_plugin import SearchPathPlugin
from hydra_plugins.anemoi_searchpath.anemoi_searchpath_plugin import AnemoiSearchPathPlugin


def test_anemoi_home_searchpath_discovery() -> None:
    # Tests that this plugin can be discovered via the plugins subsystem when looking at all Plugins
    assert AnemoiSearchPathPlugin.__name__ in [x.__name__ for x in Plugins.instance().discover(SearchPathPlugin)]


def test_config_installed() -> None:
    with initialize(version_base=None):
        config_loader = GlobalHydra.instance().config_loader()
        assert "default" in config_loader.get_group_options("hydra/output")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from contextlib import nullcontext

import pytorch_lightning as pl
import torch
from omegaconf import OmegaConf
from pytorch_lightning.callbacks import Callback

LOGGER = logging.getLogger(__name__)


class RolloutEval(Callback):
    """Evaluates the model performance over a (longer) rollout window."""

    def __init__(self, config: OmegaConf, rollout: int, every_n_batches: int) -> None:
        """Initialize RolloutEval callback.

        Parameters
        ----------
        config : dict
            Dictionary with configuration settings
        rollout : int
            Rollout length for evaluation
        every_n_batches : int
            Frequency of rollout evaluation, runs every `n` validation batches

        """
        super().__init__()
        self.config = config

        LOGGER.debug(
            "Setting up RolloutEval callback with rollout = %d, every_n_batches = %d ...",
            rollout,
            every_n_batches,
        )
        self.rollout = rollout
        self.every_n_batches = every_n_batches

    def _eval(
        self,
        pl_module: pl.LightningModule,
        batch: torch.Tensor,
    ) -> None:
        loss = torch.zeros(1, dtype=batch.dtype, device=pl_module.device, requires_grad=False)
        metrics = {}

        assert batch.shape[1] >= self.rollout + pl_module.multi_step, (
            "Batch length not sufficient for requested validation rollout length! "
            f"Set `dataloader.validation_rollout` to at least {max(self.rollout)}"
        )

        with torch.no_grad():
            for loss_next, metrics_next, _ in pl_module._rollout_step(
                batch,
                rollout=self.rollout,
                validation_mode=True,
            ):
                loss += loss_next
                metrics.update(metrics_next)

            # scale loss
            loss *= 1.0 / self.rollout
            self._log(pl_module, loss, metrics, batch.shape[0])

    def _log(self, pl_module: pl.LightningModule, loss: torch.Tensor, metrics: dict, bs: int) -> None:

        loss_scales = loss
        loss = loss_scales.sum()

        pl_module.log(
            f"val_r{self.rollout}_{getattr(pl_module.loss, 'name', pl_module.loss.__class__.__name__.lower())}",
            loss,
            on_epoch=True,
            on_step=True,
            prog_bar=False,
            logger=pl_module.logger_enabled,
            batch_size=bs,
            sync_dist=False,
            rank_zero_only=True,
        )
        if loss_scales.numel() > 1:
            for scale in range(loss_scales.numel()):
                pl_module.log(
                    f"val_r{self.rollout}_{getattr(pl_module.loss, 'name', pl_module.loss.__class__.__name__.lower())}",
                    loss_scales[scale],
                    on_epoch=True,
                    on_step=True,
                    prog_bar=False,
                    logger=pl_module.logger_enabled,
                    batch_size=bs,
                    sync_dist=False,
                    rank_zero_only=True,
                )

        for mname, mvalue in metrics.items():
            for scale in range(mvalue.numel()):

                log_val = mvalue[scale] if mvalue.numel() > 1 else mvalue

                pl_module.log(
                    f"val_r{self.rollout}_" + mname + "_scale_" + str(scale),
                    log_val,
                    on_epoch=True,
                    on_step=False,
                    prog_bar=False,
                    logger=pl_module.logger_enabled,
                    batch_size=bs,
                    sync_dist=False,
                    rank_zero_only=True,
                )

    def on_validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: list,
        batch: torch.Tensor,
        batch_idx: int,
    ) -> None:
        del outputs  # outputs are not used
        if batch_idx % self.every_n_batches == 0:
            precision_mapping = {
                "16-mixed": torch.float16,
                "bf16-mixed": torch.bfloat16,
            }
            prec = trainer.precision
            dtype = precision_mapping.get(prec)
            context = torch.autocast(device_type=batch.device.type, dtype=dtype) if dtype is not None else nullcontext()

            with context:
                self._eval(pl_module, batch)


class RolloutEvalEns(RolloutEval):
    """Evaluates the model performance over a (longer) rollout window."""

    def _eval(self, pl_module: pl.LightningModule, batch: torch.Tensor) -> None:
        """Rolls out the model and calculates the validation metrics.

        Parameters
        ----------
        pl_module : pl.LightningModule
            Lightning module object
        batch: torch.Tensor
            Batch tensor (bs, input_steps + forecast_steps, latlon, nvar)
        """
        loss = torch.zeros(pl_module.loss.num_scales, dtype=batch.dtype, device=pl_module.device, requires_grad=False)
        assert (
            batch.shape[1] >= self.rollout + pl_module.multi_step
        ), f"Batch length ({batch.shape[1]}) not sufficient for requested rollout length!"

        metrics = {}
        with torch.no_grad():
            for loss_next, metrics_next, *_ in pl_module._rollout_step(
                batch=batch,
                rollout=self.rollout,
                validation_mode=True,
            ):
                loss += loss_next
                metrics.update(metrics_next)

            # scale loss
            loss *= 1.0 / self.rollout
            self._log(pl_module, loss, metrics, batch.shape[0])

    def on_validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: list,
        batch: torch.Tensor,
        batch_idx: int,
    ) -> None:
        del outputs  # outputs are not used
        if batch_idx % self.every_n_batches == 0 and pl_module.ens_comm_group_rank == 0:
            precision_mapping = {
                "16-mixed": torch.float16,
                "bf16-mixed": torch.bfloat16,
            }
            prec = trainer.precision
            dtype = precision_mapping.get(prec)
            context = torch.autocast(device_type=batch.device.type, dtype=dtype) if dtype is not None else nullcontext()

            with context:
                self._eval(pl_module, batch)



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
import torch
from pytest_mock import MockerFixture

from anemoi.training.losses import AlmostFairKernelCRPS
from anemoi.training.losses import MSELoss
from anemoi.training.losses.base import BaseLoss
from anemoi.training.losses.multiscale import MultiscaleLossWrapper
from anemoi.training.losses.multiscale import SparseProjector


@pytest.fixture
def loss_inputs_multiscale() -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Fixture for loss inputs."""
    tensor_shape = [1, 2, 4, 1]  # (bs, time, grid, vars)

    pred = torch.zeros(tensor_shape)
    pred[0, :, 0, 0] = torch.tensor([1.0, 1.0])
    target = torch.zeros(tensor_shape[1:])

    # With only one "grid point" differing by 1 in all
    # variables, the loss should be 1.0

    loss_result = torch.tensor([1.0])
    return pred, target, loss_result


def test_multi_scale_instantiation(loss_inputs_multiscale: tuple[torch.Tensor, torch.Tensor, torch.Tensor]) -> None:
    """Test multiscale loss instantiation with single scale."""
    per_scale_loss = AlmostFairKernelCRPS()
    multiscale_loss = MultiscaleLossWrapper(
        per_scale_loss=per_scale_loss,
        weights=[1.0],
        keep_batch_sharded=False,
    )

    pred, target, loss_result = loss_inputs_multiscale
    loss = multiscale_loss(pred, target)

    assert isinstance(loss, torch.Tensor)
    assert torch.allclose(loss, loss_result), "Loss should be equal to the expected result"


@pytest.mark.parametrize("per_scale_loss", [AlmostFairKernelCRPS(), MSELoss()])
@pytest.mark.parametrize("weights", [torch.tensor([0.3, 0.7]), torch.tensor([1.0, 2.0])])
def test_multi_scale(
    loss_inputs_multiscale: tuple[torch.Tensor, torch.Tensor, torch.Tensor],
    per_scale_loss: BaseLoss,
    weights: torch.Tensor,
    mocker: MockerFixture,
) -> None:
    """Test multiscale loss with different per-scale losses and weights."""
    smoothing_matrix = SparseProjector(
        torch.tensor([[0, 0, 1, 1, 2, 2, 3, 3], [0, 1, 1, 2, 2, 3, 3, 0]]),
        torch.ones(8) / 2,
        src_size=4,
        dst_size=4,
        row_normalize=False,
        transpose=False,
    )
    mocker.patch(
        "anemoi.training.losses.multiscale.MultiscaleLossWrapper._load_smoothing_matrices",
        return_value=[None, smoothing_matrix],
    )

    multiscale_loss = MultiscaleLossWrapper(
        loss_matrices=[None, "fake"],
        per_scale_loss=per_scale_loss,
        weights=weights,
        keep_batch_sharded=False,
    )

    pred, target, _ = loss_inputs_multiscale
    loss = multiscale_loss(pred, target, squash=True)

    assert isinstance(loss, torch.Tensor)
    assert loss.shape == (2,), "Loss should have shape (num_scales,) when squash=True"

    loss = multiscale_loss(pred, target, squash=False)

    assert isinstance(loss, torch.Tensor)
    assert loss.shape == (2, pred.shape[-1]), "Loss should have shape (num_scales, num_variables) when squash=False"


def test_multiscale_loss_equivalent_to_per_scale_loss() -> None:
    """Test equivalence when only one scale is used."""
    tensor_shape = [1, 2, 4, 5]
    pred = torch.randn(tensor_shape)
    target = torch.randn(tensor_shape[1:])

    per_scale_loss = AlmostFairKernelCRPS()
    multiscale_loss = MultiscaleLossWrapper(
        per_scale_loss=per_scale_loss,
        weights=[1.0],
        keep_batch_sharded=False,
    )

    loss = multiscale_loss(pred, target)
    loss_kcrps = per_scale_loss(pred, target)

    assert isinstance(loss, torch.Tensor)
    assert torch.allclose(loss, loss_kcrps), "Loss for single/original scale should be equal to the kcrps"



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import importlib
import io
import logging
import pickle
from pathlib import Path
from typing import Any

import torch
import torch.nn as nn
from pytorch_lightning import Callback
from pytorch_lightning import LightningModule
from pytorch_lightning import Trainer

from anemoi.models.migrations import Migrator
from anemoi.training.train.tasks.base import BaseGraphModule
from anemoi.utils.checkpoints import save_metadata

chunking_fix_migration = importlib.import_module("anemoi.models.migrations.scripts.1762857428_chunking_fix").migrate

LOGGER = logging.getLogger(__name__)


def load_and_prepare_model(lightning_checkpoint_path: str) -> tuple[torch.nn.Module, dict]:
    """Load the lightning checkpoint and extract the pytorch model and its metadata.

    Parameters
    ----------
    lightning_checkpoint_path : str
        path to lightning checkpoint

    Returns
    -------
    tuple[torch.nn.Module, dict]
        pytorch model, metadata

    """
    module = BaseGraphModule.load_from_checkpoint(lightning_checkpoint_path, weights_only=False)
    model = module.model

    metadata = dict(**model.metadata)
    model.metadata = None
    model.config = None

    return model, metadata


def save_inference_checkpoint(model: torch.nn.Module, metadata: dict, save_path: Path | str) -> Path:
    """Save a pytorch checkpoint for inference with the model metadata.

    Parameters
    ----------
    model : torch.nn.Module
        Pytorch model
    metadata : dict
        Anemoi Metadata to inject into checkpoint
    save_path : Path | str
        Directory to save anemoi checkpoint

    Returns
    -------
    Path
        Path to saved checkpoint
    """
    save_path = Path(save_path)
    inference_filepath = save_path.parent / f"inference-{save_path.name}"

    torch.save(model, inference_filepath)
    save_metadata(inference_filepath, metadata)
    return inference_filepath


def transfer_learning_loading(model: torch.nn.Module, ckpt_path: Path | str) -> nn.Module:
    # Load the checkpoint
    checkpoint = torch.load(ckpt_path, weights_only=False, map_location=model.device)

    # apply chunking migration (fails silently otherwise leading to hard to debug issues)
    # this is due to loading with strict=False, planning to make this more robust in the future
    checkpoint = chunking_fix_migration(checkpoint)

    # Filter out layers with size mismatch
    state_dict = checkpoint["state_dict"]

    model_state_dict = model.state_dict()

    for key in state_dict.copy():
        if key in model_state_dict and state_dict[key].shape != model_state_dict[key].shape:
            LOGGER.info("Skipping loading parameter: %s", key)
            LOGGER.info("Checkpoint shape: %s", str(state_dict[key].shape))
            LOGGER.info("Model shape: %s", str(model_state_dict[key].shape))

            del state_dict[key]  # Remove the mismatched key

    # Load the filtered st-ate_dict into the model
    model.load_state_dict(state_dict, strict=False)
    # Needed for data indices check
    model._ckpt_model_name_to_index = checkpoint["hyper_parameters"]["data_indices"].name_to_index
    return model


def freeze_submodule_by_name(module: nn.Module, target_name: str) -> None:
    """Recursively freezes the parameters of a submodule with the specified name.

    Parameters
    ----------
    module : torch.nn.Module
        Pytorch model
    target_name : str
        The name of the submodule to freeze.
    """
    for name, child in module.named_children():
        # If this is the target submodule, freeze its parameters
        if name == target_name:
            for param in child.parameters():
                param.requires_grad = False
        else:
            # Recursively search within children
            freeze_submodule_by_name(child, target_name)


class LoggingUnpickler(pickle.Unpickler):
    def find_class(self, module: str, name: str) -> str:
        if "anemoi.training" in module:
            msg = (
                f"anemoi-training Pydantic schemas found in model's metadata: "
                f"({module}, {name}) Please review Pydantic schemas to avoid this."
            )
            raise ValueError(msg)
        return super().find_class(module, name)


def check_classes(model: torch.nn.Module) -> None:
    buffer = io.BytesIO()
    pickle.dump(model, buffer)
    buffer.seek(0)
    _ = LoggingUnpickler(buffer).load()


class RegisterMigrations(Callback):
    """Callback that register all existing migrations to a checkpoint before storing it."""

    def __init__(self):
        self.migrator = Migrator()

    def on_save_checkpoint(
        self,
        trainer: Trainer,  # noqa: ARG002
        pl_module: LightningModule,  # noqa: ARG002
        checkpoint: dict[str, Any],
    ) -> None:
        self.migrator.register_migrations(checkpoint)



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from anemoi.training.losses.combined import CombinedLoss
from anemoi.training.utils.enums import TensorDim

if TYPE_CHECKING:
    import numpy as np

    from anemoi.models.data_indices.collection import IndexCollection
    from anemoi.training.losses.base import BaseLoss

LOGGER = logging.getLogger(__name__)


def reduce_to_last_dim(x: np.ndarray) -> np.ndarray:
    if x.ndim > 1:
        return x.sum(axis=tuple(range(x.ndim - 1)))
    return x


def print_variable_scaling(loss: BaseLoss, data_indices: IndexCollection) -> dict[str, float]:
    """Log the final variable scaling for each variable in the model and return the scaling values.

    Parameters
    ----------
    loss : BaseLoss
        Loss function to get the variable scaling from.
    data_indices : IndexCollection
        Index collection to get the variable names from.

    Returns
    -------
    Dict[str, float]
        Dictionary mapping variable names to their scaling values. If max_variables is specified,
        only the top N variables plus 'total_sum' will be included.
    """
    # TODO(frazane,anaprietonem): handle special cases more gracefully
    # https://github.com/ecmwf/anemoi-core/pull/723#discussion_r2597831922
    if isinstance(loss, CombinedLoss):
        variable_scaling = {}
        for sub_loss in loss.losses:
            variable_scaling[sub_loss.__class__.__name__] = print_variable_scaling(sub_loss, data_indices)
        return variable_scaling

    variable_scaling = loss.scaler.subset_by_dim(TensorDim.VARIABLE.value).get_scaler(len(TensorDim)).reshape(-1)
    log_text = f"Final Variable Scaling in {loss.__class__.__name__}: "
    scaling_values, scaling_sum = {}, 0.0

    for idx, name in enumerate(data_indices.model.output.name_to_index.keys()):
        value = float(variable_scaling[idx])
        log_text += f"{name}: {value:.4g}, "
        scaling_values[name] = value
        scaling_sum += value

    log_text += f"Total scaling sum: {scaling_sum:.4g}, "
    scaling_values["total_sum"] = scaling_sum
    LOGGER.debug(log_text)

    return scaling_values



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import torch
from torch_geometric.data import HeteroData

from anemoi.training.utils.masks import Boolean1DMask
from anemoi.training.utils.masks import NoOutputMask


def test_boolean1d_apply() -> None:
    num_points = 10
    x = torch.reshape(torch.arange(1, num_points + 1), (1, 1, num_points, 1))
    y = -x.clone()
    mask = torch.arange(num_points) % 2 == 0
    graph = HeteroData()
    graph["nodes"]["attr"] = mask
    bool_1d = Boolean1DMask(graph, "nodes", "attr")

    x_1 = bool_1d.apply(x, dim=2, fill_value=y)

    assert x_1.shape == x.shape
    assert torch.all(x_1[:, :, ~mask, :] < 0)
    assert torch.all(x_1[:, :, ~mask, :] == y[:, :, ~mask, :])
    assert torch.all(x_1[:, :, mask, :] > 0)
    assert torch.all(x_1[:, :, mask, :] == x[:, :, mask, :])

    x_2 = bool_1d.apply(x, dim=2, fill_value=0)

    assert x_2.shape == x.shape
    assert torch.all(x_2[:, :, ~mask, :] == 0)
    assert torch.all(x_2[:, :, mask, :] > 0)
    assert torch.all(x_2[:, :, mask, :] == x[:, :, mask, :])


def test_nooutput_apply() -> None:
    num_points = 10
    x = torch.reshape(torch.arange(1, num_points + 1), (1, 1, num_points, 1))
    y = -x.clone()

    no_mask = NoOutputMask()

    x_1 = no_mask.apply(x, dim=2, fill_value=y)

    assert x_1.shape == x.shape
    assert torch.all(x_1 == x)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import torch
from torch.utils.checkpoint import checkpoint

from anemoi.models.distributed.graph import gather_tensor
from anemoi.training.train.tasks.rollout import BaseRolloutGraphModule

if TYPE_CHECKING:
    from collections.abc import Generator

    from omegaconf import DictConfig
    from torch.distributed.distributed_c10d import ProcessGroup
    from torch_geometric.data import HeteroData

LOGGER = logging.getLogger(__name__)


class GraphEnsForecaster(BaseRolloutGraphModule):
    """Graph neural network forecaster for ensembles for PyTorch Lightning."""

    def __init__(
        self,
        *,
        config: DictConfig,
        graph_data: HeteroData,
        statistics: dict,
        statistics_tendencies: dict,
        data_indices: dict,
        metadata: dict,
        supporting_arrays: dict,
    ) -> None:
        """Initialize graph neural network forecaster.

        Parameters
        ----------
        config : DictConfig
            Job configuration
        statistics : dict
            Statistics of the training data
        data_indices : dict
            Indices of the training data,
        metadata : dict
            Provenance information
        """
        super().__init__(
            config=config,
            graph_data=graph_data,
            statistics=statistics,
            statistics_tendencies=statistics_tendencies,
            data_indices=data_indices,
            metadata=metadata,
            supporting_arrays=supporting_arrays,
        )

        # num_gpus_per_ensemble >= 1 and num_gpus_per_ensemble >= num_gpus_per_model (as per the DDP strategy)
        self.model_comm_group_size = config.system.hardware.num_gpus_per_model
        num_gpus_per_model = config.system.hardware.num_gpus_per_model
        num_gpus_per_ensemble = config.system.hardware.num_gpus_per_ensemble

        assert num_gpus_per_ensemble % num_gpus_per_model == 0, (
            "Invalid ensemble vs. model size GPU group configuration: "
            f"{num_gpus_per_ensemble} mod {num_gpus_per_model} != 0.\
            If you would like to run in deterministic mode, please use aifs-train"
        )

        self.lr = (
            config.system.hardware.num_nodes
            * config.system.hardware.num_gpus_per_node
            * config.training.lr.rate
            / num_gpus_per_ensemble
        )
        LOGGER.info("Base (config) learning rate: %e -- Effective learning rate: %e", config.training.lr.rate, self.lr)

        self.nens_per_device = config.training.ensemble_size_per_device
        self.nens_per_group = self.nens_per_device * num_gpus_per_ensemble // num_gpus_per_model
        LOGGER.info("Ensemble size: per device = %d, per ens-group = %d", self.nens_per_device, self.nens_per_group)

        # lazy init ensemble group info, will be set by the DDPEnsGroupStrategy:
        self.ens_comm_group = None
        self.ens_comm_group_id = None
        self.ens_comm_group_rank = None
        self.ens_comm_num_groups = None
        self.ens_comm_group_size = None

    def set_ens_comm_group(
        self,
        ens_comm_group: ProcessGroup,
        ens_comm_group_id: int,
        ens_comm_group_rank: int,
        ens_comm_num_groups: int,
        ens_comm_group_size: int,
    ) -> None:
        self.ens_comm_group = ens_comm_group
        self.ens_comm_group_id = ens_comm_group_id
        self.ens_comm_group_rank = ens_comm_group_rank
        self.ens_comm_num_groups = ens_comm_num_groups
        self.ens_comm_group_size = ens_comm_group_size

    def set_ens_comm_subgroup(
        self,
        ens_comm_subgroup: ProcessGroup,
        ens_comm_subgroup_id: int,
        ens_comm_subgroup_rank: int,
        ens_comm_subgroup_num_groups: int,
        ens_comm_subgroup_size: int,
    ) -> None:
        self.ens_comm_subgroup = ens_comm_subgroup
        self.ens_comm_subgroup_id = ens_comm_subgroup_id
        self.ens_comm_subgroup_rank = ens_comm_subgroup_rank
        self.ens_comm_subgroup_num_groups = ens_comm_subgroup_num_groups
        self.ens_comm_subgroup_size = ens_comm_subgroup_size

    def compute_loss_metrics(
        self,
        y_pred: torch.Tensor,
        y: torch.Tensor,
        step: int | None = None,
        validation_mode: bool = False,
    ) -> tuple[torch.Tensor | None, dict[str, torch.Tensor]]:
        y_pred_ens = gather_tensor(
            y_pred.clone(),  # for bwd because we checkpoint this region
            dim=1,
            shapes=[y_pred.shape] * self.ens_comm_subgroup_size,
            mgroup=self.ens_comm_subgroup,
        )

        loss = self._compute_loss(
            y_pred_ens,
            y,
            grid_dim=self.grid_dim,
            grid_shard_shape=self.grid_shard_shapes,
        )

        # Compute metrics if in validation mode
        metrics_next = {}
        if validation_mode:
            metrics_next = self._compute_metrics(y_pred_ens, y, step=step, grid_shard_slice=self.grid_shard_slice)

        return loss, metrics_next, y_pred_ens

    def _rollout_step(
        self,
        batch: torch.Tensor,
        rollout: int | None = None,
        validation_mode: bool = False,
    ) -> Generator[tuple[torch.Tensor | None, dict, list]]:
        """Rollout step for the forecaster.

        Parameters
        ----------
        batch : torch.Tensor
            Normalized batch to use for rollout (assumed to be already preprocessed)
        rollout : int, optional
            Number of times to rollout for, by default None
            If None, will use self.rollout
        validation_mode : bool, optional
            Whether in validation mode, and to calculate validation metrics, by default False
            If False, metrics will be empty

        Yields
        ------
        Generator[tuple[torch.Tensor | None, dict, list], None, None]
            Loss value, metrics, and predictions (per step)

        Returns
        -------
        None
            None
        """
        # Stack the analysis nens_per_device times along an ensemble dimension
        x = batch[
            :,
            0 : self.multi_step,
            ...,
            self.data_indices.data.input.full,
        ]  # (bs, ms, ens_dummy, latlon, nvar)

        x = torch.cat([x] * self.nens_per_device, dim=2)  # shape == (bs, ms, nens_per_device, latlon, nvar)
        LOGGER.debug("Shapes: x.shape = %s", list(x.shape))

        assert len(x.shape) == 5, f"Expected a 5-dimensional tensor and got {len(x.shape)} dimensions, shape {x.shape}!"
        assert (x.shape[1] == self.multi_step) and (x.shape[2] == self.nens_per_device), (
            "Shape mismatch in x! "
            f"Expected ({self.multi_step}, {self.nens_per_device}), "
            f"got ({x.shape[1]}, {x.shape[2]})!"
        )

        msg = (
            "Batch length not sufficient for requested multi_step length!"
            f", {batch.shape[1]} !>= {rollout + self.multi_step}"
        )
        assert batch.shape[1] >= rollout + self.multi_step, msg

        for rollout_step in range(rollout or self.rollout):
            # prediction at rollout step rollout_step, shape = (bs, latlon, nvar)
            y_pred = self(x, fcstep=rollout_step)

            y = batch[:, self.multi_step + rollout_step, 0, :, self.data_indices.data.output.full]
            LOGGER.debug("SHAPE: y.shape = %s", list(y.shape))
            # y includes the auxiliary variables, so we must leave those out when computing the loss

            loss, metrics_next, y_pred_ens_group = checkpoint(
                self.compute_loss_metrics,
                y_pred,
                y,
                step=rollout_step,
                validation_mode=validation_mode,
                use_reentrant=False,
            )

            x = self._advance_input(x, y_pred, batch, rollout_step)

            yield loss, metrics_next, y_pred_ens_group



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import time
import uuid
from pathlib import Path

import pytorch_lightning as pl
import torch
import torchinfo
from omegaconf import OmegaConf
from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint
from pytorch_lightning.utilities import rank_zero_only

from anemoi.training.utils.checkpoint import check_classes
from anemoi.utils.checkpoints import save_metadata

LOGGER = logging.getLogger(__name__)


class AnemoiCheckpoint(ModelCheckpoint):
    """A checkpoint callback that saves the model after every validation epoch."""

    def __init__(self, config: OmegaConf, **kwargs: dict) -> None:
        """Initialise the AnemoiCheckpoint callback.

        Parameters
        ----------
        config : OmegaConf
            Config object
        kwargs : dict
            Additional keyword arguments for Pytorch ModelCheckpoint

        """
        super().__init__(**kwargs)

        self.config = config
        self.start = time.time()
        self._model_metadata = None
        self._tracker_metadata = None
        self._tracker_name = None

    @staticmethod
    def _torch_drop_down(trainer: pl.Trainer) -> torch.nn.Module:
        # Get the model from the DataParallel wrapper, for single and multi-gpu cases
        assert hasattr(trainer, "model"), "Trainer has no attribute 'model'! Is the Pytorch Lightning version correct?"
        return trainer.model.module.model if hasattr(trainer.model, "module") else trainer.model.model

    @rank_zero_only
    def model_metadata(self, model: torch.nn.Module) -> dict:
        if self._model_metadata is not None:
            return self._model_metadata

        self._model_metadata = {
            "model": model.__class__.__name__,
            "trainable_parameters": sum(p.numel() for p in model.parameters() if p.requires_grad),
            "total_parameters": sum(p.numel() for p in model.parameters()),
            "summary": repr(
                torchinfo.summary(
                    model,
                    depth=50,
                    verbose=0,
                    row_settings=["var_names"],
                ),
            ),
        }

        return self._model_metadata

    def _adjust_epoch_progress(self, trainer: pl.Trainer) -> None:
        """Adjust the epoch progress when saving a mid-epoch checkpoint.

        Since Pytorch Lightning advances one epoch at end of training (on_train-end),
        we need to correct the checkpoint epoch progress to avoid inconsistencies.
        """
        trainer.fit_loop.epoch_progress.current.processed = trainer.fit_loop.epoch_progress.current.processed - 1
        trainer.fit_loop.epoch_progress.current.completed = trainer.fit_loop.epoch_progress.current.completed - 1
        trainer.fit_loop.epoch_progress.total.processed = trainer.fit_loop.epoch_progress.total.processed - 1
        trainer.fit_loop.epoch_progress.total.completed = trainer.fit_loop.epoch_progress.total.completed - 1

    def on_train_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Save the last checkpoint at the end of training.

        If the candidates aren't better than the last checkpoint, then no checkpoints are saved.
        Note - this method if triggered when using max_epochs, it won't save any checkpoints
        since the monitor candidates won't show any changes with regard the the 'on_train_epoch_end' hook.
        """
        del pl_module
        if not self._should_skip_saving_checkpoint(trainer) and not self._should_save_on_train_epoch_end(trainer):
            if trainer.fit_loop.epoch_progress.current.completed == trainer.fit_loop.epoch_progress.current.ready:
                self._adjust_epoch_progress(trainer)
            monitor_candidates = self._monitor_candidates(trainer)
            self._save_topk_checkpoint(trainer, monitor_candidates)
            self._save_last_checkpoint(trainer, monitor_candidates)

    def tracker_metadata(self, trainer: pl.Trainer) -> dict:
        if self._tracker_metadata is not None:
            return {self._tracker_name: self._tracker_metadata}

        if self.config.diagnostics.log.wandb.enabled:
            self._tracker_name = "wand"
            import wandb

            run = wandb.run
            if run is not None:
                self._tracker_metadata = {
                    "id": run.id,
                    "name": run.name,
                    "url": run.url,
                    "project": run.project,
                }
            return {self._tracker_name: self._tracker_metadata}

        if self.config.diagnostics.log.mlflow.enabled:
            self._tracker_name = "mlflow"

            from anemoi.training.diagnostics.mlflow.logger import BaseAnemoiMLflowLogger

            mlflow_logger = next(logger for logger in trainer.loggers if isinstance(logger, BaseAnemoiMLflowLogger))
            run_id = mlflow_logger.run_id
            run = mlflow_logger._mlflow_client.get_run(run_id)

            if run is not None:
                self._tracker_metadata = {
                    "id": run.info.run_id,
                    "name": run.info.run_name,
                    "url": run.info.artifact_uri,
                    "project": run.info.experiment_id,
                }
            return {self._tracker_name: self._tracker_metadata}

        return {}

    def _remove_checkpoint(self, trainer: pl.Trainer, filepath: str) -> None:
        """Calls the strategy to remove the checkpoint file."""
        super()._remove_checkpoint(trainer, filepath)
        trainer.strategy.remove_checkpoint(self._get_inference_checkpoint_filepath(filepath))

    def _get_inference_checkpoint_filepath(self, filepath: str) -> str:
        """Defines the filepath for the inference checkpoint."""
        return Path(filepath).parent / Path("inference-" + str(Path(filepath).name))

    def on_train_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Check that model's metadata does not contain Pydantic schemas references."""
        del pl_module

        if trainer.is_global_zero:
            model = self._torch_drop_down(trainer)
            check_classes(model)

    def _save_checkpoint(self, trainer: pl.Trainer, lightning_checkpoint_filepath: str) -> None:
        if trainer.is_global_zero:
            model = self._torch_drop_down(trainer)

            # We want a different uuid each time we save the model
            # so we can tell them apart in the catalogue (i.e. different epochs)
            checkpoint_uuid = str(uuid.uuid4())
            trainer.lightning_module._hparams["metadata"]["uuid"] = checkpoint_uuid

            trainer.lightning_module._hparams["metadata"]["model"] = self.model_metadata(model)
            trainer.lightning_module._hparams["metadata"]["tracker"] = self.tracker_metadata(trainer)

            trainer.lightning_module._hparams["metadata"]["training"] = {
                "current_epoch": trainer.current_epoch,
                "global_step": trainer.global_step,
                "elapsed_time": time.time() - self.start,
            }

            Path(lightning_checkpoint_filepath).parent.mkdir(parents=True, exist_ok=True)

            save_config = model.config
            model.config = None

            tmp_metadata = model.metadata
            model.metadata = None

            tmp_supporting_arrays = model.supporting_arrays
            model.supporting_arrays = None

            # Make sure we don't accidentally modify these
            metadata = tmp_metadata.copy()
            supporting_arrays = tmp_supporting_arrays.copy()

            inference_checkpoint_filepath = self._get_inference_checkpoint_filepath(lightning_checkpoint_filepath)

            torch.save(model, inference_checkpoint_filepath)

            save_metadata(inference_checkpoint_filepath, metadata, supporting_arrays=supporting_arrays)

            model.config = save_config
            model.metadata = tmp_metadata
            model.supporting_arrays = tmp_supporting_arrays

            self._last_global_step_saved = trainer.global_step

        trainer.strategy.barrier()

        # saving checkpoint used for pytorch-lightning based training
        trainer.save_checkpoint(lightning_checkpoint_filepath, self.save_weights_only)

        self._last_global_step_saved = trainer.global_step
        self._last_checkpoint_saved = lightning_checkpoint_filepath

        if trainer.is_global_zero:
            from importlib.metadata import version
            from weakref import proxy

            from packaging.version import Version

            if Version(version("torch")) >= Version("2.6"):
                # Add a new uuid
                checkpoint_uuid = str(uuid.uuid4())
                trainer.lightning_module._hparams["metadata"]["uuid"] = checkpoint_uuid

                # Extract and save metadata for lightning checkpoint
                model = self._torch_drop_down(trainer)
                metadata = model.metadata.copy()
                supporting_arrays = model.supporting_arrays.copy()

                save_metadata(lightning_checkpoint_filepath, metadata, supporting_arrays=supporting_arrays)

            # notify loggers
            for logger in trainer.loggers:
                logger.after_save_checkpoint(proxy(self))



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from __future__ import annotations

import logging
import os
import types
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Literal

from pytorch_lightning.loggers.mlflow import _convert_params
from pytorch_lightning.utilities.rank_zero import rank_zero_only

if TYPE_CHECKING:
    from argparse import Namespace

    from mlflow.tracking import MlflowClient
    from pytorch_lightning.loggers.mlflow import MLFlowLogger

try:
    if TYPE_CHECKING:
        from azure.ai.ml.entities import Workspace

    from azure.ai.ml import MLClient
    from azure.ai.ml.identity import AzureMLOnBehalfOfCredential
    from azure.identity import DefaultAzureCredential
    from azure.identity import ManagedIdentityCredential

    # NOTE: Lightweight dependency for Python 3.10.
    # Replace with `from enum import StrEnum` when deprecating 3.10.
    from strenum import StrEnum

except ModuleNotFoundError as e:
    msg = (
        "Use of MLFlow logging in Azure requires the modules `azure-ai-ml` and `azure-identity`. You can install these"
        "via the Azure optional extra in Anemoi training: `pip install anemoi-training[azure]`."
    )
    raise ModuleNotFoundError(msg) from e


from anemoi.training.diagnostics.mlflow import LOG_MODEL
from anemoi.training.diagnostics.mlflow import MAX_PARAMS_LENGTH
from anemoi.training.diagnostics.mlflow.logger import BaseAnemoiMLflowLogger
from anemoi.utils.mlflow.auth import NoAuth

LOGGER = logging.getLogger(__name__)


class AzureIdentity(StrEnum):
    USER = "user_identity"
    MANAGED = "managed"
    DEFAULT = "default"


def get_azure_workspace(
    auth_type: str,
    subscription_id: str | None = None,
    resource_group: str | None = None,
    workspace_name: str | None = None,
) -> Workspace:
    # validate auth_type
    try:
        azure_id_type = AzureIdentity(auth_type.strip().casefold())
    except ValueError as e:
        valid_options = ", ".join(x.value for x in AzureIdentity)
        msg = f"AzureML auth type needs to be one of: {valid_options}. Recieved {auth_type}"
        raise ValueError(msg) from e

    # these env variables are usually attatched to azure ml jobs,
    # so use them if they exist.
    sub = subscription_id or os.getenv("AZUREML_ARM_SUBSCRIPTION")
    rg = resource_group or os.getenv("AZUREML_ARM_RESOURCEGROUP")
    wsname = workspace_name or os.getenv("AZUREML_ARM_WORKSPACE_NAME")

    if sub and rg and wsname:
        LOGGER.info("Attempting Azure authentication with the following details:")
        LOGGER.info("Subscription: %s", sub)
        LOGGER.info("Resource group: %s", rg)
        LOGGER.info("Workspace: %s", wsname)

        match azure_id_type:
            case AzureIdentity.MANAGED:
                client_id = os.getenv("DEFAULT_IDENTITY_CLIENT_ID")
                credential = (
                    ManagedIdentityCredential(client_id=client_id) if client_id else ManagedIdentityCredential()
                )
            case AzureIdentity.USER:
                credential = AzureMLOnBehalfOfCredential()
            case AzureIdentity.DEFAULT:
                credential = DefaultAzureCredential()

        ml_client = MLClient(
            credential,
            subscription_id=sub,
            resource_group_name=rg,
            workspace_name=wsname,
        )
        LOGGER.info("Successfully authenticated with Azure.")
    else:
        msg = (
            "Azure environment incorrectly configured; tried to use \n  "
            f"- subscription: {sub}\n  - resource_group: {rg}\n  - workspace: {wsname}.\n"
            "Try explicitly setting your subscription details via `diagnostics.mlflow.subscription_id`,"
            "`diagnostics.mlflow.resource_group`, `diagnostics.mlflow.workspace_name`."
        )
        raise ValueError(msg)
    LOGGER.info("Attempting to get Workspace object...")
    ws = ml_client.workspaces.get(ml_client.workspace_name)
    LOGGER.info("Succeeded getting the current workspace!")
    return ws


class AnemoiAzureMLflowLogger(BaseAnemoiMLflowLogger):
    """A custom MLflow logger that logs to AzureML."""

    def __init__(
        self,
        identity: AzureIdentity,
        subscription_id: str | None = None,
        resource_group: str | None = None,
        workspace_name: str | None = None,
        azure_log_level: str = "WARNING",
        experiment_name: str = "lightning_logs",
        project_name: str = "anemoi",
        run_name: str | None = None,
        tracking_uri: str | None = os.getenv("MLFLOW_TRACKING_URI"),
        save_dir: str | None = None,
        log_model: Literal["all"] | bool = LOG_MODEL,
        prefix: str = "",
        run_id: str | None = None,
        fork_run_id: str | None = None,
        offline: bool | None = False,
        authentication: bool | None = False,
        system: bool | None = True,
        terminal: bool | None = False,
        log_hyperparams: bool | None = True,
        expand_hyperparams: dict | None = None,
        on_resume_create_child: bool | None = True,
        max_params_length: int | None = MAX_PARAMS_LENGTH,
        http_max_retries: int | None = 35,
    ) -> None:
        """Initialize the AnemoiAzureMLflowLogger.

        Parameters
        ----------
        identity: str | None, optional
            Type of authentication to fall back on for accessing the AzureML workspace.
        subscription_id: str | None, optional
            The Azure subscription id
        resource_group: str | None, optional
            Name of the Azure ML resource group
        workspace: str | None, optional
            Name of the Azure ML workspace
        azure_log_level: str, optional
            Log level for all azure packages (azure-identity, azure-core, etc)
        """
        if offline:
            LOGGER.exception("Cannot run AnemoiAzureMLflowLogger with offline=True")
            raise ValueError

        if terminal:
            LOGGER.warning("Cannot log terminal output with AzureML version of MLFlow logger, will set terminal=False")

        # Set azure logging to warning, since otherwise it's way too much
        azure_logger = logging.getLogger("azure")
        numeric_level = getattr(logging, azure_log_level.upper(), None)
        if not isinstance(numeric_level, int):
            msg = f"AnemoiAzureMLflowLogger received invalid azure_log_level: {azure_log_level}"
            LOGGER.exception(msg)
            raise TypeError
        azure_logger.setLevel(numeric_level)

        # Azure ML jobs (should) set this for us:
        tracking_uri = tracking_uri or os.getenv("MLFLOW_TRACKING_URI")

        # fall back to subscription-based method if not
        if not tracking_uri:
            LOGGER.warning(
                "Could not retrieve Azure MLFlow uri automatically; trying to retrieve from subscription...",
            )
            tracking_uri = get_azure_workspace(
                identity or "default",
                subscription_id,
                resource_group,
                workspace_name,
            ).mlflow_tracking_uri

        # Azure sets the mlflow run id when the user runs az ml job create,
        # so unless we are forking/resuming etc, we want to grab this environment variable
        run_id = run_id or os.getenv("MLFLOW_RUN_ID")

        super().__init__(
            experiment_name=experiment_name,
            project_name=project_name,
            run_name=run_name,
            tracking_uri=tracking_uri,
            save_dir=save_dir,
            log_model=log_model,
            prefix=prefix,
            run_id=run_id,
            fork_run_id=fork_run_id,
            offline=offline,
            authentication=authentication,
            system=system,
            terminal=terminal,
            log_hyperparams=log_hyperparams,
            expand_hyperparams=expand_hyperparams,
            on_resume_create_child=on_resume_create_child,
            max_params_length=max_params_length,
            http_max_retries=http_max_retries,
        )

        # replace logger.experiment.log_artifact with one that checks for resource conflicts
        # note that we do this here since this is a unique issue for AzureML MLFlow loggers
        self._original_log_artifact = self.experiment.log_artifact
        self.experiment.log_artifact = types.MethodType(self._patched_log_artifact, self.experiment)

    def _init_authentication(self) -> None:
        self.auth = NoAuth()

    def _patched_log_artifact(
        self,
        _experiment_self: MLFlowLogger.experiment,  # leading underscore bypasses ruff, see note in docstring
        *args,
        **kwargs,
    ) -> None:
        """Overwrites MLFlowLogger.experiment.log_artifact.

        AzureML does not allow artifacts to be overwritten, and will raise an exception related to a resource conflict.
        To avoid this, we add the try/except logic here since this is unique to AML. Note that it's not clear **why**
        some plots are being logged more than once, but until that's understood this patch is necessary.

        Note:
            We explicitly pass experiment_self even though it's not used because it is passed to the wrapped
            log_artifact method during usage. The leading underscore bypasses the ruff check.
        """
        try:
            return self._original_log_artifact(*args, **kwargs)

        except Exception as e:
            # Ignore resource conflicts for AML
            if "Resource Conflict" in str(e) or "already exists" in str(e):
                msg = f"AnemoiAzureMLflowLogger.experiment.log_artifact: Resource conflict ignored: {e}"
                LOGGER.debug(msg)
                return None
            # Otherwise, re-raise unexpected exceptions
            raise

    @rank_zero_only
    def log_hyperparams(
        self,
        params: dict[str, Any] | Namespace,
    ) -> None:
        """Overwrite the log_hyperparams method.

        Note:
            AzureML has a strict 200 parameter limit. Since there is no easy
            way for users to limit the number of parameters they are logging,
            this method does not log each individual parameter. It only logs
            all hyperparameters together in a single .json artifact.

        Parameters
        ----------
        params : dict[str, Any] | Namespace
            params to log
        """
        if self._flag_log_hparams:
            params = _convert_params(params)

            # this is needed to resolve optional missing config values to a string, instead of raising a missing error
            if config := params.get("config"):
                params["config"] = config.model_dump(by_alias=True)

            self.log_hyperparams_as_mlflow_artifact(
                client=self.experiment,
                run_id=self.run_id,
                params=params,
            )

    @staticmethod
    def log_hyperparams_as_mlflow_artifact(
        client: MlflowClient,
        run_id: str,
        params: dict[str, Any] | Namespace,
    ) -> None:
        """Log hyperparameters as an artifact."""
        import datetime
        import json
        import tempfile
        from json import JSONEncoder

        class StrEncoder(JSONEncoder):
            def default(self, o: Any) -> str:
                return str(o)

        now = datetime.datetime.now(datetime.timezone.utc)
        now = str(now).replace(" ", "T")
        with tempfile.TemporaryDirectory() as tmp_dir:
            path = Path(tmp_dir) / f"config.{now}.json"
            with Path.open(path, "w") as f:
                json.dump(params, f, cls=StrEncoder)
            client.log_artifact(run_id=run_id, local_path=path)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import numpy as np
import pytorch_lightning as pl
import torch
from lightning_fabric.utilities.optimizer import _optimizers_to_device
from pytorch_lightning.overrides.distributed import _sync_module_states
from pytorch_lightning.strategies.ddp import DDPStrategy
from pytorch_lightning.trainer.states import TrainerFn

from anemoi.training.utils.seeding import get_base_seed

LOGGER = logging.getLogger(__name__)


def register_gradient_scaling_hooks(
    model: torch.nn.Module,
    model_comm_group_size: float,
    skip_grad_scaling: list[str] | None = None,
) -> None:
    """Register parameter hooks for gradient reduction.

    Here, we rescale parameters that only see a subset of the input on each rank
    -> these are still divided by the total number of GPUs in DDP as if each rank would see a full set of inputs
    note: the trainable parameters are added before the split across GPUs and are therefore not rescaled.

    Parameters
    ----------
    model : torch.nn.Module
        The model to register hooks on.
    model_comm_group_size : float
        The size of the model communication group for scaling.
    skip_grad_scaling : list[str] | None
        List of parameter name patterns to skip gradient scaling.
        Defaults to ["trainable", "no_gradscaling"].
    """
    if skip_grad_scaling is None:
        skip_grad_scaling = ["trainable", "no_gradscaling"]

    for name, param in model.named_parameters():
        if not param.requires_grad:
            continue
        if any(skip_name in name for skip_name in skip_grad_scaling):
            continue
        param.register_hook(lambda grad: grad * float(model_comm_group_size))


def seed_rnd(model_comm_group_id: int, global_rank: int) -> None:
    """Seed the random number generators for the rank."""
    base_seed = get_base_seed()
    initial_seed = base_seed * (model_comm_group_id + 1)
    rnd_seed = pl.seed_everything(initial_seed)  # note: workers are seeded independently in dataloader
    np_rng = np.random.default_rng(rnd_seed)
    sanity_rnd = (torch.rand(1), np_rng.random())
    LOGGER.debug(
        (
            "Strategy: Rank %d, model comm group id %d, base seed %d, seeded with %d, "
            "running with random seed: %d, sanity rnd: %s"
        ),
        global_rank,
        model_comm_group_id,
        base_seed,
        initial_seed,
        rnd_seed,
        sanity_rnd,
    )


def get_my_model_comm_group(num_gpus_per_model: int, global_rank: int, world_size: int) -> tuple[int, int, int]:
    """Determine tasks that work together and from a model group.

    Parameters
    ----------
    num_gpus_per_model : int
        Number of GPUs per model to shard over.

    Returns
    -------
    tuple[int, int, int]
        Model_comm_group id, Model_comm_group rank, Number of model_comm_groups
    """
    model_comm_group_id = global_rank // num_gpus_per_model
    model_comm_group_rank = global_rank % num_gpus_per_model
    model_comm_num_groups = world_size // num_gpus_per_model

    return model_comm_group_id, model_comm_group_rank, model_comm_num_groups


def get_my_reader_group(model_comm_group_rank: int, read_group_size: int, global_rank: int) -> tuple[int, int, int]:
    """Determine tasks that work together and from a reader group.

    Parameters
    ----------
    model_comm_group_rank : int
        Rank within the model communication group.
    read_group_size : int
        Number of dataloader readers per model group.

    Returns
    -------
    tuple[int, int, int]
        Reader_group id, Reader_group rank, Reader_group root (global rank)
    """
    reader_group_id = model_comm_group_rank // read_group_size
    reader_group_rank = model_comm_group_rank % read_group_size
    reader_group_size = read_group_size
    reader_group_root = (global_rank // read_group_size) * read_group_size

    return reader_group_id, reader_group_rank, reader_group_size, reader_group_root


class DDPGroupStrategy(DDPStrategy):
    """Distributed Data Parallel strategy with group communication."""

    def __init__(self, num_gpus_per_model: int, read_group_size: int, **kwargs: dict) -> None:
        """Initialize the distributed strategy.

        Parameters
        ----------
        num_gpus_per_model : int
            Number of GPUs per model to shard over.
        read_group_size : int
            Number of GPUs per reader group.
        **kwargs : dict
            Additional keyword arguments.

        """
        super().__init__(**kwargs)
        self.model_comm_group_size = num_gpus_per_model
        self.read_group_size = read_group_size

    def setup(self, trainer: pl.Trainer) -> None:
        assert self.accelerator is not None, "Accelerator is not initialized for distributed strategy"
        self.accelerator.setup(trainer)

        # determine the model groups that work together:

        assert self.world_size % self.model_comm_group_size == 0, (
            f"Total number of GPUs ({self.world_size}) must be divisible by the number of GPUs "
            f"per model ({self.model_comm_group_size})."
        )

        model_comm_group_ranks = np.split(
            np.arange(self.world_size, dtype=int),
            int(self.world_size / self.model_comm_group_size),
        )
        model_comm_groups = [
            torch.distributed.new_group(x) for x in model_comm_group_ranks
        ]  # every rank has to create all of these

        model_comm_group_id, model_comm_group_rank, model_comm_num_groups = get_my_model_comm_group(
            self.model_comm_group_size,
            self.global_rank,
            self.world_size,
        )
        model_comm_group = model_comm_groups[model_comm_group_id]
        self.model.set_model_comm_group(
            model_comm_group,
            model_comm_group_id,
            model_comm_group_rank,
            model_comm_num_groups,
            self.model_comm_group_size,
        )

        # set up reader groups by further splitting model_comm_group_ranks with read_group_size:

        assert self.model_comm_group_size % self.read_group_size == 0, (
            f"Number of GPUs per model ({self.model_comm_group_size}) must be divisible by read_group_size "
            f"({self.read_group_size})."
        )

        reader_group_ranks = np.array(
            [
                np.split(group_ranks, int(self.model_comm_group_size / self.read_group_size))
                for group_ranks in model_comm_group_ranks
            ],
        )  # Shape: (num_model_comm_groups, model_comm_grp_size/read_group_size, read_group_size)
        reader_groups = [[torch.distributed.new_group(x) for x in group_ranks] for group_ranks in reader_group_ranks]
        reader_group_id, reader_group_rank, reader_group_size, reader_group_root = get_my_reader_group(
            model_comm_group_rank,
            self.read_group_size,
            self.global_rank,
        )
        # get all reader groups of the current model group
        model_reader_groups = reader_groups[model_comm_group_id]
        self.model.set_reader_groups(
            model_reader_groups,
            reader_group_id,
            reader_group_rank,
            reader_group_size,
        )

        LOGGER.debug(
            "Rank %d model_comm_group_id: %d model_comm_group: %s model_comm_group_rank: %d "
            "reader_group_id: %d reader_group: %s reader_group_rank: %d reader_group_root (global): %d",
            self.global_rank,
            model_comm_group_id,
            str(model_comm_group_ranks[model_comm_group_id]),
            model_comm_group_rank,
            reader_group_id,
            reader_group_ranks[model_comm_group_id, reader_group_id],
            reader_group_rank,
            reader_group_root,
        )

        # register hooks for correct gradient reduction
        self.register_parameter_hooks()

        # move the model to the correct device
        self.model_to_device()

        # skip wrapping the model if we are not fitting as no gradients need to be exchanged
        trainer_fn = trainer.state.fn

        if trainer_fn == TrainerFn.FITTING and self._layer_sync:
            assert self.model is not None, "Model is not initialized for distributed strategy"
            self.model = self._layer_sync.apply(self.model)

        self.setup_precision_plugin()

        if trainer_fn == TrainerFn.FITTING:
            # do not wrap with DDP if not fitting as there's no gradients to reduce
            self.configure_ddp()

            # set up optimizers after the wrapped module has been moved to the device
            self.setup_optimizers(trainer)
            _optimizers_to_device(self.optimizers, self.root_device)

            import torch.distributed.algorithms.ddp_comm_hooks.post_localSGD_hook as post_localSGD

            if isinstance(self._ddp_comm_state, post_localSGD.PostLocalSGDState):
                self._enable_model_averaging()
        else:
            # we need to manually synchronize the module's states since we aren't using the DDP wrapper
            assert self.model is not None, "Model is not initialized for distributed strategy"
            _sync_module_states(self.model)

        # seed ranks
        seed_rnd(model_comm_group_id, self.global_rank)

    def process_dataloader(self, dataloader: torch.utils.data.DataLoader) -> torch.utils.data.DataLoader:
        """Pass communication group information to the dataloader for distributed training.

        Parameters
        ----------
        dataloader : torch.utils.data.DataLoader
            Dataloader to process.

        Returns
        -------
        torch.utils.data.DataLoader
            Processed dataloader.

        """
        dataloader = super().process_dataloader(dataloader)

        # pass model and reader group information to the dataloaders dataset
        model_comm_group_id, model_comm_group_rank, model_comm_num_groups = get_my_model_comm_group(
            self.model_comm_group_size,
            self.global_rank,
            self.world_size,
        )
        _, reader_group_rank, _, _ = get_my_reader_group(
            model_comm_group_rank,
            self.read_group_size,
            self.global_rank,
        )

        dataloader.dataset.set_comm_group_info(
            self.global_rank,
            model_comm_group_id,
            model_comm_group_rank,
            model_comm_num_groups,
            reader_group_rank,
            self.read_group_size,
        )

        return dataloader

    def register_parameter_hooks(self) -> None:
        """Register parameter hooks for gradient reduction."""
        register_gradient_scaling_hooks(self.model, self.model_comm_group_size)


class DDPEnsGroupStrategy(DDPStrategy):
    """Distributed Data Parallel strategy with group communication for ensembles."""

    def __init__(self, num_gpus_per_model: int, num_gpus_per_ensemble: int, read_group_size: int, **kwargs) -> None:
        """Initialize the distributed strategy.

        Parameters
        ----------
        num_gpus_per_model : int
            Number of GPUs per model to shard over.
        read_group_size : int
            Number of GPUs per reader group.
        **kwargs : dict
            Additional keyword arguments.

        """
        super().__init__(**kwargs)
        self.model_comm_group_size = num_gpus_per_model
        self.read_group_size = read_group_size
        self.ens_comm_group_size = num_gpus_per_ensemble

    def setup(self, trainer: pl.Trainer) -> None:
        assert self.accelerator is not None, "Accelerator is not initialized for distributed strategy"
        self.accelerator.setup(trainer)

        # determine the model groups that work together:

        assert self.world_size % self.model_comm_group_size == 0, (
            f"Total number of GPUs ({self.world_size}) must be divisible by the number of GPUs "
            f"per model ({self.model_comm_group_size})."
        )

        model_comm_group_ranks = np.split(
            np.arange(self.world_size, dtype=int),
            int(self.world_size / self.model_comm_group_size),
        )
        model_comm_groups = [
            torch.distributed.new_group(x) for x in model_comm_group_ranks
        ]  # every rank has to create all of these

        model_comm_group_id, model_comm_group_rank, model_comm_num_groups = get_my_model_comm_group(
            self.model_comm_group_size,
            self.global_rank,
            self.world_size,
        )
        model_comm_group = model_comm_groups[model_comm_group_id]
        self.model.set_model_comm_group(
            model_comm_group,
            model_comm_group_id,
            model_comm_group_rank,
            model_comm_num_groups,
            self.model_comm_group_size,
        )

        # set up reader groups by further splitting model_comm_group_ranks with read_group_size:

        assert self.model_comm_group_size % self.read_group_size == 0, (
            f"Number of GPUs per model ({self.model_comm_group_size}) must be divisible by read_group_size "
            f"({self.read_group_size})."
        )

        reader_group_ranks = np.array(
            [
                np.split(group_ranks, int(self.model_comm_group_size / self.read_group_size))
                for group_ranks in model_comm_group_ranks
            ],
        )  # Shape: (num_model_comm_groups, model_comm_grp_size/read_group_size, read_group_size)
        reader_groups = [[torch.distributed.new_group(x) for x in group_ranks] for group_ranks in reader_group_ranks]
        reader_group_id, reader_group_rank, reader_group_size, reader_group_root = get_my_reader_group(
            model_comm_group_rank,
            self.read_group_size,
            self.global_rank,
        )
        # get all reader groups of the current model group
        model_reader_groups = reader_groups[model_comm_group_id]
        self.model.set_reader_groups(
            model_reader_groups,
            reader_group_id,
            reader_group_rank,
            reader_group_size,
        )

        LOGGER.info(
            "Rank %d model_comm_group_id: %d model_comm_group: %s "
            "model_comm_group_rank: %d model_comm_group.size(): %d "
            "reader_group_id: %d reader_group: %s reader_group_rank: %d "
            "reader_group_root (global): %d "
            "model_reader_groups: %s reader_groups: %s",
            self.global_rank,
            model_comm_group_id,
            str(model_comm_group_ranks[model_comm_group_id]),
            model_comm_group_rank,
            model_comm_group.size(),
            reader_group_id,
            reader_group_ranks[model_comm_group_id, reader_group_id],
            reader_group_rank,
            reader_group_root,
            model_reader_groups,
            reader_groups,
        )

        # determine the ensemble groups that work together:
        assert self.world_size % self.ens_comm_group_size == 0, (
            f"Total number of GPUs ({self.world_size}) must be divisible by the number of GPUs "
            f"per ensemble ({self.ens_comm_group_size})."
        )
        assert self.ens_comm_group_size % self.model_comm_group_size == 0, (
            f"Number of GPUs per ensemble ({self.ens_comm_group_size}) must be divisible by the number of GPUs "
            f"per model ({self.model_comm_group_size})."
        )

        ens_comm_group_ranks = np.split(
            np.arange(self.world_size, dtype=int),
            int(self.world_size / self.ens_comm_group_size),
        )
        ens_comm_groups = [torch.distributed.new_group(x) for x in ens_comm_group_ranks]

        ens_comm_group_id, ens_comm_group_rank, ens_comm_num_groups = get_my_model_comm_group(
            self.ens_comm_group_size,
            self.global_rank,
            self.world_size,
        )

        ens_comm_group = ens_comm_groups[ens_comm_group_id]
        self.model.set_ens_comm_group(
            ens_comm_group,
            ens_comm_group_id,
            ens_comm_group_rank,
            ens_comm_num_groups,
            self.ens_comm_group_size,
        )

        # ens_comm_subgroup: subgroup of same model_comm_group ranks inside the ensemble group
        spacing = self.model_comm_group_size
        ens_comm_subgroup_ranks = [
            ens_comm_group[offset::spacing] for ens_comm_group in ens_comm_group_ranks for offset in range(spacing)
        ]

        ens_comm_subgroups = [torch.distributed.new_group(x) for x in ens_comm_subgroup_ranks]

        ens_comm_subgroup_size = self.ens_comm_group_size // self.model_comm_group_size
        ens_comm_subgroup_id = ens_comm_group_id * self.model_comm_group_size + model_comm_group_rank
        ens_comm_subgroup_rank = ens_comm_group_rank // self.model_comm_group_size
        ens_comm_num_subgroups = self.world_size // ens_comm_subgroup_size

        ens_comm_subgroup = ens_comm_subgroups[ens_comm_subgroup_id]
        self.model.set_ens_comm_subgroup(
            ens_comm_subgroup,
            ens_comm_subgroup_id,
            ens_comm_subgroup_rank,
            ens_comm_num_subgroups,
            ens_comm_subgroup_size,
        )

        LOGGER.info(
            "Rank %d ens_comm_group_id: %d ens_comm_group: %s ens_comm_group_rank: %d "
            "ens_comm_group_size: %d ens_comm_group.size(): %d ens_comm_subgroup_id: %d "
            "ens_comm_subgroup: %s ens_comm_subgroup_rank: %d ens_comm_subgroup.size(): %d ",
            self.global_rank,
            ens_comm_group_id,
            str(ens_comm_group_ranks[ens_comm_group_id]),
            ens_comm_group_rank,
            self.ens_comm_group_size,
            ens_comm_group.size(),
            ens_comm_subgroup_id,
            str(ens_comm_subgroup_ranks[ens_comm_subgroup_id]),
            ens_comm_subgroup_rank,
            ens_comm_subgroup_size,
        )

        # register hooks for correct gradient reduction
        self.register_parameter_hooks()

        # move the model to the correct device
        self.model_to_device()

        # skip wrapping the model if we are not fitting as no gradients need to be exchanged
        trainer_fn = trainer.state.fn

        if trainer_fn == TrainerFn.FITTING and self._layer_sync:
            assert self.model is not None, "Model is not initialized for distributed strategy"
            self.model = self._layer_sync.apply(self.model)

        self.setup_precision_plugin()

        if trainer_fn == TrainerFn.FITTING:
            # do not wrap with DDP if not fitting as there's no gradients to reduce
            self.configure_ddp()

            # set up optimizers after the wrapped module has been moved to the device
            self.setup_optimizers(trainer)
            _optimizers_to_device(self.optimizers, self.root_device)

            import torch.distributed.algorithms.ddp_comm_hooks.post_localSGD_hook as post_localSGD

            if isinstance(self._ddp_comm_state, post_localSGD.PostLocalSGDState):
                self._enable_model_averaging()
        else:
            # we need to manually synchronize the module's states since we aren't using the DDP wrapper
            assert self.model is not None, "Model is not initialized for distributed strategy"
            _sync_module_states(self.model)

        # seed ranks
        seed_rnd(model_comm_group_id, self.global_rank)

    def process_dataloader(self, dataloader: torch.utils.data.DataLoader) -> torch.utils.data.DataLoader:
        """Pass communication group information to the dataloader for distributed training.

        Parameters
        ----------
        dataloader : torch.utils.data.DataLoader
            Dataloader to process.

        Returns
        -------
        torch.utils.data.DataLoader
            Processed dataloader.

        """
        dataloader = super().process_dataloader(dataloader)

        # pass model and reader group information to the dataloaders dataset
        model_comm_group_id, model_comm_group_rank, model_comm_num_groups = get_my_model_comm_group(
            self.model_comm_group_size,
            self.global_rank,
            self.world_size,
        )
        _, reader_group_rank, _, _ = get_my_reader_group(
            model_comm_group_rank,
            self.read_group_size,
            self.global_rank,
        )
        ens_comm_group_id, ens_comm_group_rank, ens_comm_num_groups = get_my_model_comm_group(
            self.ens_comm_group_size,
            self.global_rank,
            self.world_size,
        )

        dataloader.dataset.set_comm_group_info(
            self.global_rank,
            model_comm_group_id,
            model_comm_group_rank,
            model_comm_num_groups,
            reader_group_rank,
            self.read_group_size,
        )

        dataloader.dataset.set_ens_comm_group_info(
            ens_comm_group_id,
            ens_comm_group_rank,
            ens_comm_num_groups,
        )

        return dataloader

    def register_parameter_hooks(self) -> None:
        """Register parameter hooks for gradient reduction."""
        register_gradient_scaling_hooks(self.model, self.model_comm_group_size)
